"""
Protun protocol implementation.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=R0801 # Duplicate code with wireguard.py and openvpn.py but that's expected for now.

from __future__ import annotations

import json
import socket
import uuid
import logging
from concurrent.futures import Future
from datetime import datetime
from getpass import getuser
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path
from typing import Dict, Optional, Union, List

import gi

gi.require_version("NM", "1.0")  # noqa: required before importing NM module
# pylint: disable=wrong-import-position
from gi.repository import NM

from proton.vpn.connection import events, states
from proton.vpn.connection.events import EventContext
from proton.vpn.connection.interfaces import Settings
from proton.vpn.backend.networkmanager.core import LinuxNetworkManager, LocalAgentMixin

from proton.vpn.core.settings.packet_capture import PacketCaptureMode

try:
    import proton.vpn.linux.protun as proton_vpn_linux_protun
except ImportError:
    # The protun plugin might not be available in the environment,
    # so we handle the import error gracefully.
    proton_vpn_linux_protun = None


logger = logging.getLogger(__name__)

SERVICE_TYPE = "org.freedesktop.NetworkManager.protun"
STORE_PRIVATE_KEY_IN_KEYRING = "1"
PRIVATE_KEY = "private-key"
PRIVATE_KEY_FLAGS = "private-key-flags"

# Internal network configuration — same WireGuard infrastructure as wireguard.py
_INTERNAL_IPV4_ADDRESS = "10.2.0.2"
_INTERNAL_IPV4_PREFIX = 32
_INTERNAL_IPV4_DNS = "10.2.0.1"

_INTERNAL_IPV6_ADDRESS = "2a07:b944::2:2"
_INTERNAL_IPV6_PREFIX = 128
_INTERNAL_IPV6_DNS = "2a07:b944::2:1"

_DNS_SEARCH = "~"
_DNS_PRIORITY = -1500


def get_dns(ip_version: Union[type[IPv4Address], type[IPv6Address]]):
    """
    Get the correct dns server address depending on whether we are using
    ipv6 or ipv4.
    """
    if ip_version == IPv4Address:
        return _INTERNAL_IPV4_DNS
    return _INTERNAL_IPV6_DNS


def generate_capture_path(directory_path: str,
                          now: Optional[datetime] = None) -> Path:
    """
    Generate a file path for a Proton VPN packet capture file.

    Args:
        directory_path (str or Path): The base directory path
        now (datetime, optional): The datetime to use for the filename.
                                  Defaults to datetime.now() if not provided.
                                  Useful for unit testing.

    Returns:
        Path: Path object pointing to the capture file with timestamp and .pcap extension
    """
    base_dir = Path(directory_path)

    if now is None:
        now = datetime.now()

    date_str = now.strftime(r"%Y_%m_%d")
    time_str = now.strftime(r"%H_%M_%S")
    filename = f"proton_vpn__{date_str}__{time_str}.pcap"
    full_path = base_dir / filename

    return full_path


class Protun(LinuxNetworkManager, LocalAgentMixin):
    """Creates a protun VPN plugin connection."""

    SIGNAL_NAME: str = "vpn-state-changed"
    VIRTUAL_DEVICE_NAME: str = "proton0"
    PLUGIN_NAME: str = "protun"
    connection: Optional[NM.SimpleConnection] = None
    plugin_exists: Optional[bool] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        LocalAgentMixin.__init__(self, self._user_tier)
        self._connection_settings = None
        self._protun_client = None
        self._protun = kwargs.get("protun", proton_vpn_linux_protun)

    def setup(self) -> Future:
        """Creates and registers the NM VPN connection."""
        self._generate_connection()
        self._modify_connection()
        return self.nm_client.add_connection_async(self.connection)

    def _generate_connection(self):
        self._unique_id = str(uuid.uuid4())
        self._connection_settings = NM.SettingConnection.new()
        self.connection = NM.SimpleConnection.new()

    def _modify_connection(self):
        self._set_custom_connection_id()
        self._set_uuid()
        self._set_interface_name()
        self._set_connection_type()
        self._set_connection_user_owned()
        self.connection.add_setting(self._connection_settings)

        self._set_route()
        self._set_dns()
        self._set_vpn_settings()

        self.connection.verify()

    def _set_custom_connection_id(self):
        self._connection_settings.set_property(NM.SETTING_CONNECTION_ID, self._get_servername())

    def _set_uuid(self):
        self._connection_settings.set_property(NM.SETTING_CONNECTION_UUID, self._unique_id)

    def _set_interface_name(self):
        self._connection_settings.set_property(
            NM.SETTING_CONNECTION_INTERFACE_NAME, self.VIRTUAL_DEVICE_NAME
        )

    def _set_connection_type(self):
        self._connection_settings.set_property(
            NM.SETTING_CONNECTION_TYPE, NM.SETTING_VPN_SETTING_NAME
        )

    def _set_connection_user_owned(self):
        self._connection_settings.add_permission(
            NM.SETTING_USER_SETTING_NAME,
            getuser(),
            None
        )

    def _set_route(self):
        self.connection.add_setting(
            self._set_route_ipv4(NM.SettingIP4Config.new()))
        self.connection.add_setting(
            self._set_route_ipv6(NM.SettingIP6Config.new()))

    def _set_route_ipv4(self, ipv4_config: NM.SettingIP4Config):
        # We set SETTING_IP4_CONFIG_METHOD_AUTO so that the network manager
        # sets up automatic routing to the VPN server IP, to avoid routing
        # loops.
        #
        # Later we'll want to replace this with fwmark which will allow protun
        # to work with split tunneling.
        ipv4_config.set_property(NM.SETTING_IP_CONFIG_METHOD,
                                 NM.SETTING_IP4_CONFIG_METHOD_AUTO)
        ipv4_config.add_address(
            NM.IPAddress.new(socket.AF_INET,
                             _INTERNAL_IPV4_ADDRESS,
                             _INTERNAL_IPV4_PREFIX)
        )

        return ipv4_config

    def _set_route_ipv6(self, ipv6_config: NM.SettingIP6Config):
        if self.enable_ipv6_support:
            ipv6_config.set_property(
                NM.SETTING_IP_CONFIG_METHOD,
                NM.SETTING_IP6_CONFIG_METHOD_MANUAL
            )
            ipv6_config.add_address(
                NM.IPAddress.new(
                    socket.AF_INET6,
                    _INTERNAL_IPV6_ADDRESS,
                    _INTERNAL_IPV6_PREFIX
                )
            )
            # Route all IPv6 traffic through our interface.
            # Unlike IPv4, there is no routing loop risk because the server is
            # reached over IPv4.
            #
            # If we dont add this, all ipv6 traffic will go through the ipv6
            # killswitch.
            ipv6_config.add_route(
                NM.IPRoute.new(socket.AF_INET6, "::", 0, None, 1)
            )
        else:
            ipv6_config.set_property(
                NM.SETTING_IP_CONFIG_METHOD,
                NM.SETTING_IP6_CONFIG_METHOD_DISABLED
            )

        return ipv6_config

    def _set_dns(self):
        ipv4_config = self.connection.get_setting_ip4_config()
        ipv6_config = self.connection.get_setting_ip6_config()

        self._configure_dns(nm_setting=ipv4_config, ip_version=IPv4Address)
        if self.enable_ipv6_support:
            self._configure_dns(nm_setting=ipv6_config, ip_version=IPv6Address)

        self.connection.add_setting(ipv4_config)
        self.connection.add_setting(ipv6_config)

    def _configure_dns(
        self,
        nm_setting: Union[NM.SettingIP4Config, NM.SettingIP6Config],
        ip_version: Union[type[IPv4Address], type[IPv6Address]],
        dns_priority: int = _DNS_PRIORITY,
    ):
        """Sets DNS, respecting custom DNS settings."""
        if ip_version not in [IPv4Address, IPv6Address]:
            raise ValueError(f"Unknown IP version: {ip_version}")

        nm_setting.set_property(NM.SETTING_IP_CONFIG_DNS_PRIORITY, dns_priority)
        nm_setting.set_property(NM.SETTING_IP_CONFIG_IGNORE_AUTO_DNS, True)

        # pylint: disable=duplicate-code
        custom_dns_ips = self._settings.custom_dns\
            .get_enabled_dns_list_based_on_ip_version(ip_version)
        ip_addresses = [dns.exploded for dns in custom_dns_ips]

        # Protun does not auto-configure DNS, so we always set it explicitly.
        if self._settings.custom_dns.enabled and ip_addresses:
            nm_setting.set_property(NM.SETTING_IP_CONFIG_DNS, ip_addresses)
        else:
            nm_setting.add_dns(get_dns(ip_version))
            nm_setting.add_dns_search(_DNS_SEARCH)

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        raise NotImplementedError(
            "This method should be implemented by ProtunUDP, ProtunTCP, etc. subclasses.")

    def _set_vpn_settings(self):
        vpn_settings = NM.SettingVpn.new()
        vpn_settings.set_property(NM.SETTING_VPN_SERVICE_TYPE, SERVICE_TYPE)

        # Build the peers array expected by the protun plugin (settings.rs)
        peer = {
            "id": self._vpnserver.server_name or self._vpnserver.server_ip,
            "endpoint": f"{self._vpnserver.server_ip}",
            "public-key": self._vpnserver.x25519pk,
            "udp-ports": [],
            "tcp-ports": [],
            "tls-ports": [],
            "priority": 0,
        }
        peer.update(self._protun_ports())

        settings_str = json.dumps({
            "version": 1,
            "peers": [peer],
            "pcap-file": None
        }).replace(",", r"\,")  # Escape commas for NetworkManager

        vpn_settings.add_data_item("settings", settings_str)

        # The WireGuard private key is stored as a VPN secret.
        # NM passes it to the protun auth-dialog which forwards it to the plugin.
        vpn_settings.add_secret(
            PRIVATE_KEY,
            self._vpncredentials.pubkey_credentials.wg_private_key
        )

        # Use the keyring to store the connection private key.
        vpn_settings.add_data_item(PRIVATE_KEY_FLAGS,
                                   STORE_PRIVATE_KEY_IN_KEYRING)

        self.connection.add_setting(vpn_settings)

    # pylint: disable=arguments-renamed
    def _on_state_changed(
            self, vpn_connection: NM.VpnConnection, state: int, reason: int
    ):
        """
        Handle VPN plugin state changes emitted by NetworkManager.

        NM emits vpn-state-changed with state and reason whenever the protun
        plugin connection state changes. This translates NM states into
        backend-agnostic state machine events.
        """
        try:
            state = NM.VpnConnectionState(state)
        except ValueError:
            logger.warning("Unexpected VPN connection state: %s", state)
            state = NM.VpnConnectionState.UNKNOWN

        try:
            reason = NM.VpnConnectionStateReason(reason)
        except ValueError:
            logger.warning("Unexpected VPN connection state reason: %s", reason)
            reason = NM.VpnConnectionStateReason.UNKNOWN

        logger.debug(
            "Protun connection state changed: state=%s, reason=%s",
            state.value_name, reason.value_name
        )

        if state is NM.VpnConnectionState.ACTIVATED:
            logger.info("Starting local agent for domain: %s", self._vpnserver.domain)
            self._async_start_local_agent_listener()
            self._notify_subscribers_threadsafe(
                events.Connected(EventContext(connection=self))
            )
        elif state is NM.VpnConnectionState.FAILED:
            if reason in [
                NM.VpnConnectionStateReason.CONNECT_TIMEOUT,
                NM.VpnConnectionStateReason.SERVICE_START_TIMEOUT,
            ]:
                self._notify_subscribers_threadsafe(
                    events.Timeout(EventContext(connection=self, reason=reason))
                )
            elif reason in [
                NM.VpnConnectionStateReason.NO_SECRETS,
                NM.VpnConnectionStateReason.LOGIN_FAILED,
            ]:
                self._notify_subscribers_threadsafe(
                    events.AuthDenied(EventContext(connection=self, reason=reason))
                )
            else:
                self._notify_subscribers_threadsafe(
                    events.UnexpectedError(EventContext(connection=self, reason=reason))
                )
        elif state is NM.VpnConnectionState.DISCONNECTED:
            if reason is NM.VpnConnectionStateReason.USER_DISCONNECTED:
                self._async_stop_local_agent_listener()
                self._notify_subscribers_threadsafe(
                    events.Disconnected(EventContext(connection=self, reason=reason))
                )
            elif reason is NM.VpnConnectionStateReason.DEVICE_DISCONNECTED:
                self._async_stop_local_agent_listener()
                self._notify_subscribers_threadsafe(
                    events.DeviceDisconnected(EventContext(connection=self, reason=reason))
                )
            else:
                self._notify_subscribers_threadsafe(
                    events.UnexpectedError(EventContext(connection=self, reason=reason))
                )
        else:
            logger.debug("Ignoring VPN state change: %s", state.value_name)

    def _initialize_persisted_connection(
            self, connection_id: str
    ) -> states.State:
        """Initialize from a persisted connection, restarting the local agent if connected."""
        state = super()._initialize_persisted_connection(connection_id)

        if isinstance(state, states.Connected):
            self._async_start_local_agent_listener()
        return state

    async def update_credentials(self, credentials):
        """Notifies the VPN server that the certificate needs a refresh."""
        await super().update_credentials(credentials)
        await self._start_local_agent_listener()

    @property
    def are_feature_updates_applied_when_active(self) -> bool:
        """
        Returns whether connection feature updates are applied on the fly
        while the connection is already active, without restarting it.
        """
        return True

    async def update_settings(self, settings: Settings):
        """Update features on the active agent connection."""
        await super().update_settings(settings)
        if self._agent_listener.is_running:
            await self._request_connection_features(settings.features)

    @classmethod
    def validate(cls):
        if cls.plugin_exists is None:
            cls.plugin_exists = cls._plugin_exists(cls.PLUGIN_NAME)
        return cls.plugin_exists

    async def _get_protun_client(self):
        """Returns the cached protun client, creating it on first call."""
        if self._protun is None:
            raise RuntimeError("Protun module is not available")
        if self._protun_client is None:
            self._protun_client = await self._protun.ConnectionManager.new()
        return self._protun_client

    @classmethod
    def supports_packet_capture(cls, protun=proton_vpn_linux_protun):  # pylint: disable=W0613
        """Returns False and should be overriden by Protun protocols that support packet capture"""
        return False

    async def start_packet_capture(self):
        """Starts a packet capture session, writing to a timestamped .pcap file."""
        # Translate the mode
        if self.settings.packet_capture.mode == PacketCaptureMode.OVERWRITE:
            mode = self._protun.FileWriteMode.Overwrite
        else:
            mode = self._protun.FileWriteMode.Append

        file_path = generate_capture_path(
            self.settings.packet_capture.directory_path)

        await (await self._get_protun_client()).run(
            self._protun.Command.PcapStart(
                self._protun.PcapStart(
                    file_info=self._protun.PcapFileInfo.from_path(
                        path=str(file_path),
                        mode=mode
                    ),
                    max_bytes=self.settings.packet_capture.max_bytes
                )
            )
        )

    async def stop_packet_capture(self):
        """Stops the active packet capture session."""
        await (await self._get_protun_client()).run(
            self._protun.Command.PcapStop(self._protun.PcapStop())
        )


class ProtunUDP(Protun):
    """Protun UDP protocol implementation."""

    protocol = "protun-udp"
    ui_protocol = "Wireguard UDP"

    @classmethod
    def get_priority(cls):
        """Returns the priority of the implementation. Lower values indicate higher priority."""
        return 2

    @classmethod
    def get_key(cls):
        """Returns the protocol name."""
        return cls.protocol

    @classmethod
    def get_protocol_group(cls) -> str:
        """Returns the protocol group."""
        return "protun"

    @classmethod
    def supports_packet_capture(cls, protun=proton_vpn_linux_protun):
        """Returns True if protun supports packet capture and is available."""
        return protun is not None

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        return {
            "udp-ports": self._vpnserver.wireguard_ports.udp
        }


class ProtunTCP(Protun):
    """Protun TCP protocol implementation."""

    protocol = "protun-tcp"
    ui_protocol = "Wireguard TCP"

    @classmethod
    def get_priority(cls):
        """Returns the priority of the implementation. Lower values indicate higher priority."""
        return 3

    @classmethod
    def get_key(cls):
        """Returns the protocol name."""
        return cls.protocol

    @classmethod
    def get_protocol_group(cls) -> str:
        """Returns the protocol group."""
        return "protun"

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        return {
            "tcp-ports": self._vpnserver.wireguard_ports.tcp
        }


class ProtunTLS(Protun):
    """Protun TLS protocol implementation."""

    protocol = "protun-tls"
    ui_protocol = "Stealth"

    @classmethod
    def get_priority(cls):
        """Returns the priority of the implementation. Lower values indicate higher priority."""
        return 4

    @classmethod
    def get_key(cls):
        """Returns the protocol name."""
        return cls.protocol

    @classmethod
    def get_protocol_group(cls) -> str:
        """Returns the protocol group."""
        return "protun"

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        return {
            "tls-ports": self._vpnserver.wireguard_ports.tls
        }


class ProtunSmart(Protun):
    """
    Protun protocol implementation, automatically selecting the best
    available transport protocol.
    """

    protocol = "protun-smart"
    ui_protocol = "Smart"

    @classmethod
    def get_priority(cls):
        """Returns the priority of the implementation. Lower values indicate higher priority."""
        return 1

    @classmethod
    def get_key(cls):
        """Returns the protocol name."""
        return cls.protocol

    @classmethod
    def get_protocol_group(cls) -> str:
        """Returns the protocol group."""
        return "protun"

    def _protun_ports(self) -> Dict[str, List[int]]:
        """Returns the protun ports as a dict."""
        return {
            "udp-ports": self._vpnserver.wireguard_ports.udp,
            "tcp-ports": self._vpnserver.wireguard_ports.tcp,
            "tls-ports": self._vpnserver.wireguard_ports.tls
        }


class ProtunAuto(ProtunSmart):
    """
    Protun protocol implementation, legacy, renamed to ProtunSmart.
    TODO (ltitley): Remove this before we go to stable.
    """

    protocol = "protun-auto"
    ui_protocol = "Auto [Legacy renamed to smart]"

    @classmethod
    def get_priority(cls):
        """
        Returns the priority of the implementation. Lower values indicate a
        higher priority.
        """
        return 10



"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import tempfile

import pytest

from proton.vpn.core.settings.packet_capture import (
    PacketCapture, PacketCaptureMode,
    DEFAULT_DIRECTORY_PATH, DEFAULT_MODE, MAX_PACKET_CAPTURE_FILE_SIZE,
)


def test_default_directory_path():
    assert DEFAULT_DIRECTORY_PATH == tempfile.gettempdir()


def test_default_values():
    pc = PacketCapture.default()
    assert pc.directory_path == DEFAULT_DIRECTORY_PATH
    assert pc.mode == PacketCaptureMode.OVERWRITE
    assert pc.max_bytes == MAX_PACKET_CAPTURE_FILE_SIZE



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock, AsyncMock, call

import pytest

from proton.vpn.core.refresher import VPNDataRefresher


@pytest.mark.asyncio
async def test_enable_schedules_all_refreshers_if_the_vpn_session_is_already_loaded():
    session_holder = Mock()
    scheduler = Mock()
    client_config_refresher = Mock()
    client_config_refresher.initial_refresh_delay = 0
    server_list_refresher = Mock()
    server_list_refresher.initial_refresh_delay = 0
    certificate_refresher = Mock()
    certificate_refresher.initial_refresh_delay = 0
    feature_flag_refresher = Mock()
    feature_flag_refresher.initial_refresh_delay = 0
    notifications_refresher = Mock()
    notifications_refresher.initial_refresh_delay = 0
    refresher = VPNDataRefresher(
        session_holder=session_holder,
        scheduler=scheduler,
        client_config_refresher=client_config_refresher,
        server_list_refresher=server_list_refresher,
        certificate_refresher=certificate_refresher,
        feature_flags_refresher=feature_flag_refresher,
        notifications_refresher=notifications_refresher
    )

    session_holder.session.loaded = True

    await refresher.enable()

    assert scheduler.mock_calls == [
        call.run_after(client_config_refresher.initial_refresh_delay, client_config_refresher.refresh),
        call.run_after(server_list_refresher.initial_refresh_delay, server_list_refresher.refresh),
        call.run_after(certificate_refresher.initial_refresh_delay, certificate_refresher.refresh),
        call.run_after(feature_flag_refresher.initial_refresh_delay, feature_flag_refresher.refresh),
        call.run_after(notifications_refresher.initial_refresh_delay, notifications_refresher.refresh),
        call.start()
    ]


@pytest.mark.asyncio
async def test_enable_fetches_vpn_session_when_not_loaded_and_then_schedules_refreshers():
    session_holder = Mock()
    scheduler = Mock()
    client_config_refresher = Mock()
    client_config_refresher.initial_refresh_delay = 0
    server_list_refresher = Mock()
    server_list_refresher.initial_refresh_delay = 0
    certificate_refresher = Mock()
    certificate_refresher.initial_refresh_delay = 0
    feature_flag_refresher = Mock()
    feature_flag_refresher.initial_refresh_delay = 0
    notifications_refresher = Mock()
    notifications_refresher.initial_refresh_delay = 0

    mock_manager = Mock()
    mock_manager.session_holder = session_holder
    mock_manager.scheduler = scheduler
    mock_manager.client_config_refresher = client_config_refresher
    mock_manager.server_list_refresher = server_list_refresher
    mock_manager.certificate_refresher = certificate_refresher
    mock_manager.feature_flag_refresher = feature_flag_refresher
    mock_manager.notifications_refresher = notifications_refresher

    refresher = VPNDataRefresher(
        session_holder=session_holder,
        scheduler=scheduler,
        client_config_refresher=client_config_refresher,
        server_list_refresher=server_list_refresher,
        certificate_refresher=certificate_refresher,
        feature_flags_refresher=feature_flag_refresher,
        notifications_refresher=notifications_refresher
    )

    session_holder.session.loaded = False
    session_holder.session.fetch_session_data = AsyncMock()

    await refresher.enable()

    assert mock_manager.mock_calls == [
        call.session_holder.session.fetch_session_data(),
        call.scheduler.run_after(client_config_refresher.initial_refresh_delay, client_config_refresher.refresh),
        call.scheduler.run_after(server_list_refresher.initial_refresh_delay, server_list_refresher.refresh),
        call.scheduler.run_after(certificate_refresher.initial_refresh_delay, certificate_refresher.refresh),
        call.scheduler.run_after(feature_flag_refresher.initial_refresh_delay, feature_flag_refresher.refresh),
        call.scheduler.run_after(notifications_refresher.initial_refresh_delay, notifications_refresher.refresh),
        call.scheduler.start()
    ]



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock
import copy
import itertools

import pytest

from proton.vpn.core.settings import Settings, SettingsPersistence, NetShield, SplitTunnelingMode
from proton.vpn.killswitch.interface import KillSwitchState

FREE_TIER = 0
PLUS_TIER = 1


@pytest.fixture
def default_free_settings_dict():
    return {
        "protocol": "wireguard",
        "killswitch": KillSwitchState.OFF.value,
        "custom_dns": {
            "enabled": False,
            "ip_list": []
        },
        "ipv6": True,
        "anonymous_crash_reports": True,
        "features": {
            "netshield": NetShield.NO_BLOCK.value,
            "moderate_nat": False,
            "vpn_accelerator": True,
            "port_forwarding": False,
            "split_tunneling": {
                "enabled": False,
                "mode": SplitTunnelingMode.EXCLUDE.value,
                "config_by_mode": {
                    SplitTunnelingMode.EXCLUDE.value: {
                        "mode": SplitTunnelingMode.EXCLUDE.value,
                        "app_paths": [],
                        "ip_ranges": []
                    },
                    SplitTunnelingMode.INCLUDE.value: {
                        "mode": SplitTunnelingMode.INCLUDE.value,
                        "app_paths": [],
                        "ip_ranges": []
                    },
                }
            },
        },
    }


def test_settings_get_default(default_free_settings_dict):
    free_settings = Settings.default(FREE_TIER)

    assert free_settings.to_dict() == default_free_settings_dict


def test_settings_save_to_disk(default_free_settings_dict):
    free_settings = Settings.default(FREE_TIER)
    cache_handler_mock = Mock()

    sp = SettingsPersistence(cache_handler_mock)
    sp.save(free_settings)

    cache_handler_mock.save.assert_called_once_with(free_settings.to_dict())


def test_settings_persistence_get_returns_default_settings_and_does_not_persist_them(default_free_settings_dict):
    cache_handler_mock = Mock()
    cache_handler_mock.load.return_value = None
    sp = SettingsPersistence(cache_handler_mock)

    sp.get(FREE_TIER)

    cache_handler_mock.save.assert_not_called()


def test_settings_persistence_save_persisted_settings(default_free_settings_dict):
    cache_handler_mock = Mock()
    sp = SettingsPersistence(cache_handler_mock)

    sp.save(Settings.from_dict(default_free_settings_dict, FREE_TIER))

    cache_handler_mock.save.assert_called()


def test_settings_persistence_get_returns_in_memory_settings_if_they_were_already_loaded(default_free_settings_dict):
    cache_handler_mock = Mock()
    cache_handler_mock.load.return_value = default_free_settings_dict
    sp = SettingsPersistence(cache_handler_mock)

    sp.get(FREE_TIER)

    # The persistend settings should be loaded once, not twice.
    cache_handler_mock.load.assert_called_once()


@pytest.mark.parametrize("user_tier", [FREE_TIER, PLUS_TIER])
def test_settings_persistence_ensure_features_are_loaded_with_default_values_based_on_user_tier(user_tier):
    cache_handler_mock = Mock()
    cache_handler_mock.load.return_value = None
    sp = SettingsPersistence(cache_handler_mock)

    settings = sp.get(user_tier)

    if user_tier == FREE_TIER:
        assert settings.features.netshield == NetShield.NO_BLOCK.value
    else:
        assert settings.features.netshield == NetShield.BLOCK_MALICIOUS_URL.value


def test_settings_persistence_delete_removes_persisted_settings(default_free_settings_dict):
    cache_handler_mock = Mock()
    cache_handler_mock.load.return_value = default_free_settings_dict
    sp = SettingsPersistence(cache_handler_mock)

    sp.get(FREE_TIER)

    sp.delete()

    cache_handler_mock.remove.assert_called_once()


def test_get_ipv4_custom_dns_ips_returns_only_valid_ips(default_free_settings_dict):
    valid_ips = [
        {"ip": "1.1.1.1"},
        {"ip": "2.2.2.2"},
        {"ip": "3.3.3.3"}
    ]
    invalid_ips = [
        {"ip": "asdasd"},
        {"ip": "wasd2.q212.123123"},
        {"ip": "123123123.123123123.123123123.123123"},
        {"ip": "ef0e:e1d4:87f9:a578:5e52:fb88:46a7:010a"}
    ]
    default_free_settings_dict["custom_dns"]["ip_list"] = list(itertools.chain.from_iterable([valid_ips, invalid_ips]))
    sp = Settings.from_dict(default_free_settings_dict, FREE_TIER)
    list_of_ipv4_addresses_in_string_form = [ip.exploded for ip in sp.custom_dns.get_enabled_ipv4_ips()]

    assert [dns["ip"] for dns in valid_ips] == list_of_ipv4_addresses_in_string_form


def test_get_ipv6_custom_dns_ips_returns_only_valid_ips(default_free_settings_dict):
    valid_ips = [
        {"ip": "ef0e:e1d4:87f9:a578:5e52:fb88:46a7:010a"},
        {"ip": "0275:ef68:faeb:736b:49af:36f7:1620:9308"},
        {"ip": "4e69:39c4:9c55:5b26:7fa7:730e:4012:48b6"}
    ]
    invalid_ips = [
        {"ip": "asdasd"},
        {"ip": "wasd2.q212.123123"},
        {"ip": "1.1.1.1"},
        {"ip": "2.2.2.2"},
        {"ip": "3.3.3.3"},
        {"ip": "123123123.123123123.123123123.123123"}
    ]
    default_free_settings_dict["custom_dns"]["ip_list"] = list(itertools.chain.from_iterable([valid_ips, invalid_ips]))
    sp = Settings.from_dict(default_free_settings_dict, FREE_TIER)
    list_of_ipv6_addresses_in_string_form = [ip.exploded for ip in sp.custom_dns.get_enabled_ipv6_ips()]

    assert [dns["ip"] for dns in valid_ips] == list_of_ipv6_addresses_in_string_form


def test_settings_default_split_tunneling_data_is_built_when_config_by_mode_value_is_invalid(default_free_settings_dict):
    deep_copy_default_free_settings_dict = copy.deepcopy(default_free_settings_dict)
    default_free_settings_dict["features"]["split_tunneling"]["config_by_mode"] = {}
    free_settings = Settings.default(FREE_TIER)

    free_settings.to_dict() == deep_copy_default_free_settings_dict



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from http import HTTPStatus
from unittest.mock import Mock, AsyncMock

import pytest

from proton.vpn.core.refresher.certificate_refresher import CertificateRefresher, generate_backoff_value
from proton.vpn.core.refresher.scheduler import RunAgain
from proton.session.exceptions import ProtonAPIError


@pytest.mark.asyncio
async def test_refresh_schedules_next_refresh_if_certificate_is_expired_and_api_error_exception_is_raised():
    session_holder = Mock()
    session = session_holder.session

    refresher = CertificateRefresher(session_holder=session_holder)

    session.fetch_certificate = AsyncMock(side_effect=ProtonAPIError(
        http_code=HTTPStatus.TOO_MANY_REQUESTS,
        http_headers={},
        json_data={"Code": HTTPStatus.TOO_MANY_REQUESTS, "Error": "Error message"}
    ))
    new_certificate = Mock()
    new_certificate.remaining_time_to_next_refresh = 600

    next_refresh_delay = await refresher.refresh()


@pytest.mark.asyncio
async def test_refresh_fetches_certificate_if_expired_and_returns_next_refresh_delay():
    session_holder = Mock()
    session = session_holder.session

    refresher = CertificateRefresher(session_holder=session_holder)

    session.fetch_certificate = AsyncMock()
    new_certificate = Mock()
    new_certificate.remaining_time_to_next_refresh = 600
    session.fetch_certificate.return_value= new_certificate

    next_refresh_delay = await refresher.refresh()

    assert next_refresh_delay == RunAgain.after_seconds(new_certificate.remaining_time_to_next_refresh)


@pytest.mark.parametrize("nth_failed_attempt, expected_backoff", [
    (0, 1),
    (1, 2),
    (2, 4),
    (3, 8),
    (4, 16),
    (5, 32)
])
def test_generate_backoff_value_generates_expected_value(nth_failed_attempt, expected_backoff):
    backoff_in_seconds = 1
    random_component = 1

    backoff = generate_backoff_value(
        number_of_failed_refresh_attempts=nth_failed_attempt,
        backoff_in_seconds=backoff_in_seconds,
        random_component=random_component
    )

    assert backoff == expected_backoff



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import tempfile
from os.path import basename
from unittest.mock import AsyncMock, patch, Mock

import pytest
from proton.session.transports import TransportFactory

from proton.vpn.session import VPNSession
from proton.vpn.session.dataclasses import BugReportForm
from proton.vpn.session.dataclasses.notifications.nps_survey_response import NPSSurveyResponse

MOCK_ISP = "Proton ISP"
MOCK_COUNTRY = "Middle Earth"


def create_mock_vpn_account():
    vpn_account = Mock
    vpn_account.location = Mock()
    vpn_account.location.ISP = MOCK_ISP
    vpn_account.location.Country = MOCK_COUNTRY
    return vpn_account


@pytest.mark.asyncio
async def test_submit_report():
    s = VPNSession()
    s._vpn_account = create_mock_vpn_account()
    attachments = []

    with tempfile.NamedTemporaryFile(mode="rb") as attachment1, tempfile.NamedTemporaryFile(mode="rb") as attachment2:
        attachments.append(attachment1)
        attachments.append(attachment2)

        bug_report = BugReportForm(
            username="test_user",
            email="email@pm.me",
            title="This is a title example",
            description="This is a description example",
            client_version="1.0.0",
            client="Example",
            attachments=attachments
        )

        with patch.object(s, "async_api_request") as patched_async_api_request:
            await s.submit_bug_report(bug_report)

            patched_async_api_request.assert_called_once()
            api_request_kwargs = patched_async_api_request.call_args.kwargs

        assert api_request_kwargs["endpoint"] == s.BUG_REPORT_ENDPOINT

        submitted_data = api_request_kwargs["data"]

        assert len(submitted_data.fields) == 13

        form_field = submitted_data.fields[0]
        assert form_field.name == "OS"
        assert form_field.value == bug_report.os

        form_field = submitted_data.fields[1]
        assert form_field.name == "OSVersion"
        assert form_field.value == bug_report.os_version

        form_field = submitted_data.fields[2]
        assert form_field.name == "Client"
        assert form_field.value == bug_report.client

        form_field = submitted_data.fields[3]
        assert form_field.name == "ClientVersion"
        assert form_field.value == bug_report.client_version

        form_field = submitted_data.fields[4]
        assert form_field.name == "ClientType"
        assert form_field.value == bug_report.client_type

        form_field = submitted_data.fields[5]
        assert form_field.name == "Title"
        assert form_field.value == bug_report.title

        form_field = submitted_data.fields[6]
        assert form_field.name == "Description"
        assert form_field.value == bug_report.description

        form_field = submitted_data.fields[7]
        assert form_field.name == "Username"
        assert form_field.value == bug_report.username

        form_field = submitted_data.fields[8]
        assert form_field.name == "Email"
        assert form_field.value == bug_report.email

        form_field = submitted_data.fields[9]
        assert form_field.name == "ISP"
        assert form_field.value == MOCK_ISP

        form_field = submitted_data.fields[10]
        assert form_field.name == "Country"
        assert form_field.value == MOCK_COUNTRY

        form_field = submitted_data.fields[11]
        assert form_field.name == "Attachment-0"
        assert form_field.value == bug_report.attachments[0]
        assert form_field.filename == basename(form_field.value.name)

        form_field = submitted_data.fields[12]
        assert form_field.name == "Attachment-1"
        assert form_field.value == bug_report.attachments[1]
        assert form_field.filename == basename(form_field.value.name)


# ---------------------------------------------------------------------------
# submit_nps_response
# ---------------------------------------------------------------------------

@pytest.fixture
def mock_transport():
    transport = Mock()
    transport.async_api_request = AsyncMock(return_value={"Code": 1000})
    return transport


@pytest.fixture
def nps_session(mock_transport):
    s = VPNSession()
    s.transport_factory = TransportFactory(lambda _: mock_transport)
    s._vpn_account = create_mock_vpn_account()
    return s


@pytest.mark.asyncio
async def test_submit_nps_response_submit_uses_submit_endpoint(nps_session, mock_transport):
    response = NPSSurveyResponse(
        user_score=9,
        user_comments="Great service",
        response_type=NPSSurveyResponse.ResponseType.SUBMIT,
    )
    await nps_session.submit_nps_response(response)

    endpoint = mock_transport.async_api_request.call_args.args[0]
    assert endpoint == VPNSession.NPS_SURVEY_SUBMIT_ENDPOINT


@pytest.mark.asyncio
async def test_submit_nps_response_submit_sends_score_and_comment(nps_session, mock_transport):
    response = NPSSurveyResponse(
        user_score=8,
        user_comments="Very good",
        response_type=NPSSurveyResponse.ResponseType.SUBMIT,
    )
    await nps_session.submit_nps_response(response)

    jsondata = mock_transport.async_api_request.call_args.args[1]
    assert jsondata["Score"] == 8
    assert jsondata["Comment"] == "Very good"


@pytest.mark.asyncio
async def test_submit_nps_response_dismiss_uses_dismiss_endpoint(nps_session, mock_transport):
    response = NPSSurveyResponse(user_score=0, user_comments="")
    await nps_session.submit_nps_response(response)

    endpoint = mock_transport.async_api_request.call_args.args[0]
    assert endpoint == VPNSession.NPS_SURVEY_DISMISS_ENDPOINT


@pytest.mark.asyncio
async def test_submit_nps_response_dismiss_sends_empty_data(nps_session, mock_transport):
    response = NPSSurveyResponse(user_score=0, user_comments="")
    await nps_session.submit_nps_response(response)

    jsondata = mock_transport.async_api_request.call_args.args[1]
    assert jsondata == {}


@pytest.mark.asyncio
async def test_submit_nps_response_includes_country_header(nps_session, mock_transport):
    response = NPSSurveyResponse(user_score=5, user_comments="OK")
    await nps_session.submit_nps_response(response)

    additional_headers = mock_transport.async_api_request.call_args.args[3]
    assert additional_headers["x-pm-country"] == MOCK_COUNTRY


@pytest.mark.asyncio
async def test_submit_nps_response_uses_post_method(nps_session, mock_transport):
    response = NPSSurveyResponse()
    await nps_session.submit_nps_response(response)

    method = mock_transport.async_api_request.call_args.args[4]
    assert method == "post"




"""
Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import pytest
from dataclasses import asdict
from proton.vpn.session.dataclasses import VPNLocation


@pytest.fixture
def vpnlocation_data():
    return {
        "IP": "192.168.0.1",
        "Country": "Switzerland",
        "ISP": "SwissRandomProvider",
        "Long": 7.4474,
        "Lat": 46.9480
    }


def test_vpnlocation_deserializes_expected_dict_keys(vpnlocation_data):
    vpnlocation = VPNLocation.from_dict(vpnlocation_data)

    assert asdict(vpnlocation) == vpnlocation_data


def test_vpnlocation_deserialize_should_not_crash_with_unexpected_dict_keys(vpnlocation_data):
    vpnlocation_data["unexpected_keyword"] = "keyword and data"

    VPNLocation.from_dict(vpnlocation_data)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock, AsyncMock, call, PropertyMock, patch
import pytest

from proton.vpn.backend.networkmanager.killswitch.wireguard import WGKillSwitch
from proton.vpn.backend.networkmanager.killswitch.wireguard.killswitch_connection_handler import KillSwitchConnectionHandler


@pytest.fixture
def vpn_server():
    vpn_server_mock = Mock()
    vpn_server_mock.server_ip = "1.1.1.1"

    return vpn_server_mock


@pytest.mark.asyncio
async def test_enable_without_vpn_server_adds_ks_connection():
    ks_handler_mock = AsyncMock()
    wg_ks = WGKillSwitch(ks_handler_mock)

    await wg_ks.enable()

    assert ks_handler_mock.method_calls == [
        call.add_kill_switch_connection(False)
    ]


@pytest.mark.asyncio
async def test_enable_with_vpn_server_adds_ks_connection_and_route_for_server(vpn_server):
    ks_handler_mock = AsyncMock()
    wg_ks = WGKillSwitch(ks_handler_mock)

    await wg_ks.enable(vpn_server)

    assert ks_handler_mock.method_calls == [
        call.add_kill_switch_connection(False),
        call.add_vpn_server_route(server_ip=vpn_server.server_ip)
    ]


@pytest.mark.asyncio
async def test_disable_killswitch_removes_full_and_route_for_server():
    ks_handler_mock = AsyncMock()
    wg_ks = WGKillSwitch(ks_handler_mock)

    await wg_ks.disable()

    assert ks_handler_mock.method_calls == [
        call.remove_killswitch_connection(),
        call.remove_vpn_server_route()
    ]


@pytest.mark.asyncio
async def test_enable_ipv6_leak_protection_adds_ipv6_ks():
    ks_handler_mock = AsyncMock()

    wg_ks = WGKillSwitch(ks_handler_mock)
    await wg_ks.enable_ipv6_leak_protection()

    assert ks_handler_mock.method_calls == [
        call.add_ipv6_leak_protection()
    ]


@pytest.mark.asyncio
async def test_disable_ipv6_leak_protection_removes_ipv6_ks():
    ks_handler_mock = AsyncMock()

    wg_ks = WGKillSwitch(ks_handler_mock)
    await wg_ks.disable_ipv6_leak_protection()

    assert ks_handler_mock.method_calls == [
        call.remove_ipv6_leak_protection()
    ]


@pytest.fixture
def monkey_patch_connection_handler():
    original = KillSwitchConnectionHandler.is_network_manager_running
    KillSwitchConnectionHandler.is_network_manager_running = PropertyMock(return_value=True)
    yield KillSwitchConnectionHandler
    KillSwitchConnectionHandler.is_network_manager_running = original


@pytest.mark.parametrize("validate_params_dict, assert_bool", [
    (None, False),
    ({}, False),
    ({"protocol": "openvpn"}, False),
    ({"protocol": "wireguard"}, True)
])
@patch("proton.vpn.backend.networkmanager.killswitch.wireguard.wgkillswitch.subprocess")
def test_backend_validate(mock_subprocess, validate_params_dict, assert_bool, monkey_patch_connection_handler):
    assert WGKillSwitch._validate(validate_params_dict) == assert_bool



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
from unittest.mock import Mock, patch

import pytest

from proton.vpn.connection import VPNConnection, states
from proton.vpn.connection.persistence import ConnectionPersistence, ConnectionParameters
from proton.vpn.connection.states import StateContext
from proton.vpn.connection.interfaces import VPNServer, ProtocolPorts
from proton.vpn.session.servers.types import TierEnum

from .common import MockVpnCredentials


@pytest.fixture
def settings():
    return Mock()


@pytest.fixture
def vpn_credentials():
    return MockVpnCredentials()


@pytest.fixture
def vpn_server():
    return VPNServer(
        server_ip="10.10.1.1",
        domain="com.test-domain.www",
        x25519pk="wg_public_key",
        openvpn_ports=ProtocolPorts(
            tcp=[80, 1194], udp=[445, 5995], tls=[8080]),
        wireguard_ports=ProtocolPorts(
            tcp=[443, 88], udp=[445], tls=[8080]),
        server_name="TestServer#10",
        server_id="OYB-3pMQQA2Z2Qnp5s5nIvTVO2...lRjxhx9DCAUM9uXfM2ZUFjzPXw==",
        has_ipv6_support=False,
        label="0"
    )


@pytest.fixture
def connection_persistence_mock():
    return Mock(ConnectionPersistence)


class DummyVPNConnection(VPNConnection):
    """Dummy VPN connection implementing all the required abstract methods."""
    backend = "dummy"
    protocol = "protocol"

    def __init__(self, *args, connection_persistence = None, **kwargs):
        self.initialize_persisted_connection_mock = Mock(return_value=states.Connected(StateContext(connection=self)))

        # Make sure we don't trigger connection persistence.
        connection_persistence = connection_persistence or Mock()

        super().__init__(*args, connection_persistence=connection_persistence, user_tier=TierEnum.PLUS, **kwargs)

    def _initialize_persisted_connection(
            self, persisted_parameters: ConnectionParameters
    ) -> states.State:
        return self.initialize_persisted_connection_mock(persisted_parameters)

    def start(self):
        pass

    def stop(self):
        pass

    def refresh_certificate(self):
        pass

    def _get_connection(self):
        return None

    @classmethod
    def validate(cls) -> bool:
        return True

    @classmethod
    def get_priority(cls) -> int:
        return 100

    @classmethod
    def get_key(cls) -> str:
        return cls.protocol

    @classmethod
    def get_protocol_group(cls) -> str:
        return "dummy"


class InvalidVPNConnection(VPNConnection):
    """VPN connection class missing abstract method implementations."""

    backend = "invalid"
    protocol = "protocol"


def test_vpn_connection_subclass_raises_type_exception_if_abstract_methods_were_not_implemented():
    with pytest.raises(TypeError, match="Can't instantiate abstract class"):
        InvalidVPNConnection(server=None, credentials=None)


def test_vpn_connection_initialized_without_a_persisted_connection():
    """
    When a VPNConnection object is created without passing persisted parameters
    then it should be initialized without a unique id and with the Disconnected
    initial state.
    """
    vpnconn = DummyVPNConnection(
        server=None,
        credentials=None,
        settings=None,
        connection_id=None
    )

    assert vpnconn._unique_id is None
    vpnconn.initialize_persisted_connection_mock.assert_not_called()
    assert isinstance(vpnconn.initial_state, states.Disconnected)


@pytest.mark.asyncio
async def test_add_persistence(vpn_server, vpn_credentials, settings, connection_persistence_mock):
    vpnconn = DummyVPNConnection(
        vpn_server,
        vpn_credentials,
        settings=settings,
        connection_persistence=connection_persistence_mock,
    )
    vpnconn._unique_id = "add-persistence"
    await vpnconn.add_persistence()

    connection_persistence_mock.save.assert_called_once()
    persistence_params = connection_persistence_mock.save.call_args.args[0]
    assert persistence_params.connection_id == "add-persistence"
    assert persistence_params.backend == vpnconn.backend
    assert persistence_params.protocol == vpnconn.protocol
    assert persistence_params.server == vpn_server


@pytest.mark.asyncio
async def test_remove_persistence(vpn_server, vpn_credentials, settings, connection_persistence_mock):
    vpnconn = DummyVPNConnection(
        vpn_server,
        vpn_credentials,
        settings,
        connection_persistence=connection_persistence_mock
    )
    vpnconn._unique_id = "remove-persistence"

    await vpnconn.remove_persistence()

    connection_persistence_mock.remove.assert_called()


def test_register_subscriber_delegates_to_publisher():
    publisher_mock = Mock()
    vpnconn = DummyVPNConnection(
        server=None, credentials=None, settings=None, publisher=publisher_mock
    )

    def subscriber(event):
        pass
    vpnconn.register(subscriber)

    publisher_mock.register.assert_called_with(subscriber)


def test_unregister_subscriber_delegates_to_publisher():
    publisher_mock = Mock()
    vpnconn = DummyVPNConnection(
        server=None, credentials=None, settings=None, publisher=publisher_mock
    )

    def subscriber(event):
        pass
    vpnconn.unregister(subscriber)

    publisher_mock.unregister.assert_called_with(subscriber)


def test_get_user_pass(vpn_server, vpn_credentials, settings):
    vpnconn = DummyVPNConnection(vpn_server, vpn_credentials, settings)
    u, p = vpn_credentials.userpass_credentials.username, vpn_credentials.userpass_credentials.password
    user, password = vpnconn._get_user_pass()
    assert u == user and p == password


def test_get_user_with_default_feature_flags(vpn_server, vpn_credentials, settings):
    vpnconn = DummyVPNConnection(vpn_server, vpn_credentials, settings)
    u = vpn_credentials.userpass_credentials.username
    user, _ = vpnconn._get_user_pass(True)
    _u = "+".join([u] + vpnconn._get_feature_flags())
    assert user == _u


@pytest.mark.parametrize(
    "ns, accel, pf, rn, sf",
    [
        ("f1", False, True, False, True),
        ("f2", False, True, False, True),
        ("f3", False, True, False, True),
        ("f1", True, False, True, False),
        ("f2", True, False, True, False),
        ("f3", True, False, True, False),
    ]
)
def test_get_user_with_features(vpn_server, vpn_credentials, ns, accel, pf, rn, sf):
    from proton.vpn.connection.interfaces import Features

    class MockFeatures(Features):

        @property
        def netshield(self):
            return ns

        @property
        def vpn_accelerator(self):
            return accel

        @property
        def port_forwarding(self):
            return pf

        @property
        def moderate_nat(self):
            return rn

        @property
        def safe_mode(self):
            return sf

    settings = Mock()
    settings.features = MockFeatures()

    vpnconn = DummyVPNConnection(vpn_server, vpn_credentials, settings)
    u = vpn_credentials.userpass_credentials.username
    user, _ = vpnconn._get_user_pass(True)
    _u = "+".join([u] + vpnconn._get_feature_flags())

    assert user == _u



"""
Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from dataclasses import asdict
import pytest
from proton.vpn.session.dataclasses import VPNCertificate


@pytest.fixture
def vpncertificate_data():
    return {
        "SerialNumber": "asd879hnna!as",
        "ClientKeyFingerprint": "fingerprint",
        "ClientKey": "as243sdfs4",
        "Certificate": "certificate",
        "ExpirationTime": 123456789,
        "RefreshTime": 123456789,
        "Mode": "on",
        "DeviceName": "mock-device",
        "ServerPublicKeyMode": "mock-mode",
        "ServerPublicKey": "mock-key"
    }


def test_vpncertificate_deserializes_expected_dict_keys(vpncertificate_data):
    vpncertificate = VPNCertificate.from_dict(vpncertificate_data)

    assert asdict(vpncertificate) == vpncertificate_data


def test_vpncertificate_deserialize_should_not_crash_with_unexpected_dict_keys(vpncertificate_data):
    vpncertificate_data["unexpected_keyword"] = "keyword and data"
    VPNCertificate.from_dict(vpncertificate_data)



"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from datetime import datetime
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock

import pytest

from proton.vpn.backend.networkmanager.protocol.protun.protun import (
    Protun, ProtunUDP, generate_capture_path,
)
from proton.vpn.core.settings.packet_capture import PacketCaptureMode


# ─── generate_capture_path ────────────────────────────────────────────────────

def test_generate_capture_path():
    result = generate_capture_path("/tmp", datetime(2026, 4, 30, 14, 30, 45))
    assert result == Path("/tmp/proton_vpn__2026_04_30__14_30_45.pcap")

# ─── Protun class methods ─────────────────────────────────────────────────────

@pytest.fixture(autouse=True)
def reset_plugin_exists():
    original = Protun.plugin_exists
    yield
    Protun.plugin_exists = original


def test_supports_packet_capture_():
    assert Protun.supports_packet_capture(MagicMock()) is False


def test_udp_supports_packet_capture():
    assert ProtunUDP.supports_packet_capture(MagicMock()) is True


def test_supports_packet_capture_returns_false_when_module_unavailable():
    assert Protun.supports_packet_capture(None) is False


def test_validate_returns_false_when_plugin_does_not_exist():
    Protun.plugin_exists = False
    assert Protun.validate() is False


def test_validate_returns_true_when_plugin_exists_and_module_available():
    Protun.plugin_exists = True
    assert Protun.validate() is True


# ─── Helpers ─────────────────────────────────────────────────────────────────

def _make_instance(mode, directory_path="/tmp", max_bytes=512 * 1024 * 1024, protun=None):
    """Build a bare ProtunUDP instance without invoking __init__."""
    instance = object.__new__(ProtunUDP)
    instance._protun_client = None
    instance._protun = protun

    packet_capture = MagicMock()
    packet_capture.mode = mode
    packet_capture.directory_path = directory_path
    packet_capture.max_bytes = max_bytes

    settings = MagicMock()
    settings.packet_capture = packet_capture
    instance._settings = settings
    return instance


# ─── start_packet_capture ────────────────────────────────────────────────────

@pytest.mark.asyncio
async def test_start_packet_capture_uses_overwrite_mode():
    mock_protun = MagicMock()
    mock_protun.ConnectionManager.new = AsyncMock(return_value=MagicMock(run=AsyncMock()))
    instance = _make_instance(PacketCaptureMode.OVERWRITE, protun=mock_protun)

    await instance.start_packet_capture()

    _, kwargs = mock_protun.PcapFileInfo.from_path.call_args
    assert kwargs["mode"] is mock_protun.FileWriteMode.Overwrite


@pytest.mark.asyncio
async def test_start_packet_capture_uses_append_mode():
    mock_protun = MagicMock()
    mock_protun.ConnectionManager.new = AsyncMock(return_value=MagicMock(run=AsyncMock()))
    instance = _make_instance(PacketCaptureMode.APPEND, protun=mock_protun)

    await instance.start_packet_capture()

    _, kwargs = mock_protun.PcapFileInfo.from_path.call_args
    assert kwargs["mode"] is mock_protun.FileWriteMode.Append


@pytest.mark.asyncio
async def test_start_packet_capture_passes_max_bytes():
    mock_protun = MagicMock()
    mock_protun.ConnectionManager.new = AsyncMock(return_value=MagicMock(run=AsyncMock()))
    max_bytes = 1024 * 1024 * 100
    instance = _make_instance(PacketCaptureMode.OVERWRITE, max_bytes=max_bytes, protun=mock_protun)

    await instance.start_packet_capture()

    _, kwargs = mock_protun.PcapStart.call_args
    assert kwargs["max_bytes"] == max_bytes


@pytest.mark.asyncio
async def test_start_packet_capture_sends_pcap_start_command():
    mock_protun = MagicMock()
    mock_client = MagicMock(run=AsyncMock())
    mock_protun.ConnectionManager.new = AsyncMock(return_value=mock_client)
    instance = _make_instance(PacketCaptureMode.OVERWRITE, protun=mock_protun)

    await instance.start_packet_capture()

    mock_client.run.assert_called_once_with(
        mock_protun.Command.PcapStart.return_value
    )


# ─── stop_packet_capture ─────────────────────────────────────────────────────

@pytest.mark.asyncio
async def test_stop_packet_capture_sends_pcap_stop_command():
    mock_protun = MagicMock()
    mock_client = MagicMock(run=AsyncMock())
    mock_protun.ConnectionManager.new = AsyncMock(return_value=mock_client)
    instance = object.__new__(ProtunUDP)
    instance._protun_client = None
    instance._protun = mock_protun

    await instance.stop_packet_capture()

    mock_protun.PcapStop.assert_called_once_with()
    mock_client.run.assert_called_once_with(
        mock_protun.Command.PcapStop.return_value
    )



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import pytest

from proton.vpn.session.exceptions import ClientConfigDecodeError
from proton.vpn.session.session import ClientConfig
import time

EXPIRATION_TIME = time.time()

@pytest.fixture
def apidata():
    return {
        "Code": 1000,
        "DefaultPorts": {
            "OpenVPN": {
                "UDP": [80, 51820, 4569, 1194, 5060],
                "TCP": [443, 7770, 8443]
            },
            "WireGuard": {
                "UDP": [443, 88, 1224, 51820, 500, 4500],
                "TCP": [443],
            }
        },
        "HolesIPs": ["62.112.9.168", "104.245.144.186"],
        "ServerRefreshInterval": 10,
        "FeatureFlags": {
            "NetShield": True,
            "GuestHoles": False,
            "ServerRefresh": True,
            "StreamingServicesLogos": True,
            "PortForwarding": True,
            "ModerateNAT": True,
            "SafeMode": False,
            "StartConnectOnBoot": True,
            "PollNotificationAPI": True,
            "VpnAccelerator": True,
            "SmartReconnect": True,
            "PromoCode": False,
            "WireGuardTls": True,
            "Telemetry": True,
            "NetShieldStats": True
        },
        "SmartProtocol": {
            "OpenVPN": True,
            "IKEv2": True,
            "WireGuard": True,
            "WireGuardTCP": True,
            "WireGuardTLS": True
        },
        "RatingSettings": {
            "EligiblePlans": [],
            "SuccessConnections": 3,
            "DaysLastReviewPassed": 100,
            "DaysConnected": 3,
            "DaysFromFirstConnection": 14
        },
        "ExpirationTime": EXPIRATION_TIME
    }


def test_from_dict(apidata):
    client_config = ClientConfig.from_dict(apidata)

    assert client_config.openvpn_ports.udp == apidata["DefaultPorts"]["OpenVPN"]["UDP"]
    assert client_config.openvpn_ports.tcp == apidata["DefaultPorts"]["OpenVPN"]["TCP"]
    assert client_config.wireguard_ports.udp == apidata["DefaultPorts"]["WireGuard"]["UDP"]
    assert client_config.wireguard_ports.tcp == apidata["DefaultPorts"]["WireGuard"]["TCP"]
    assert client_config.holes_ips == apidata["HolesIPs"]
    assert client_config.server_refresh_interval == apidata["ServerRefreshInterval"]
    assert client_config.expiration_time == EXPIRATION_TIME


def test_from_dict_raises_error_when_dict_does_not_have_expected_keys():
    with pytest.raises(ClientConfigDecodeError):
        ClientConfig.from_dict({})



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import pathlib

import pytest
from proton.vpn.session import VPNSession, VPNPubkeyCredentials
from proton.vpn.session.fetcher import (
    VPNCertificate, VPNSessions, VPNSettings
)
from proton.vpn.session.credentials import VPNSecrets
from proton.vpn.session.dataclasses import VPNLocation
from proton.vpn.session.certificates import Certificate
from proton.vpn.session.exceptions import (
    VPNCertificateExpiredError, VPNCertificateFingerprintError,
    VPNCertificateError
)

DATA_DIR = pathlib.Path(__file__).parent.absolute() / 'data'
with open(DATA_DIR / 'api_cert_response.json', 'r') as f:
    VPN_CERTIFICATE_API_RESPONSE = json.load(f)
    del VPN_CERTIFICATE_API_RESPONSE["Code"]
with open(DATA_DIR / 'api_vpnsettings_response.json', 'r') as f:
    VPN_API_RESPONSE = json.load(f)
    del VPN_API_RESPONSE["Code"]
with open(DATA_DIR / 'api_vpnsessions_response.json', 'r') as f:
    VPN_SESSIONS_API_RESPONSE = json.load(f)
    del VPN_SESSIONS_API_RESPONSE["Code"]
with open(DATA_DIR / 'api_vpn_location_response.json', 'r') as f:
    VPN_LOCATION_API_RESPONSE = json.load(f)
    del VPN_LOCATION_API_RESPONSE["Code"]
with open(DATA_DIR / 'vpn_secrets.json', 'r') as f:
    VPN_SECRETS_DICT = json.load(f)


class TestVpnAccountSerialize:

    def test_fingerprints(self):
        # Check if our fingerprints are matching for secrets, API and Certificate
        # Get fingerprint from the secrets. Wireguard private key from the API is in ED25519 FORMAT ?
        private_key = VPN_SECRETS_DICT["ed25519_privatekey"]
        vpn_secrets = VPNSecrets(private_key)

        fingerprint_from_secrets = vpn_secrets.proton_fingerprint_from_x25519_pk
        # Get fingerprint from API
        fingerprint_from_api = VPN_CERTIFICATE_API_RESPONSE["ClientKeyFingerprint"]
        # Get fingerprint from Certificate
        certificate = Certificate(cert_pem=VPN_CERTIFICATE_API_RESPONSE["Certificate"])
        fingerprint_from_certificate = certificate.proton_fingerprint
        assert fingerprint_from_api == fingerprint_from_certificate
        assert fingerprint_from_secrets == fingerprint_from_certificate

    def test_vpnaccount_from_dict(self):
        vpnaccount = VPNSettings.from_dict(VPN_API_RESPONSE)
        assert vpnaccount.VPN.Name == "test"
        assert vpnaccount.VPN.Password == "passwordtest"
    
    def test_vpnaccount_to_dict(self):
        assert VPNSettings.from_dict(VPN_API_RESPONSE).to_dict() == VPN_API_RESPONSE

    def test_vpncertificate_from_dict(self):
        cert = VPNCertificate.from_dict(VPN_CERTIFICATE_API_RESPONSE)
        assert cert.SerialNumber == VPN_CERTIFICATE_API_RESPONSE["SerialNumber"]
        assert cert.ClientKeyFingerprint == VPN_CERTIFICATE_API_RESPONSE["ClientKeyFingerprint"]
        assert cert.ClientKey == VPN_CERTIFICATE_API_RESPONSE["ClientKey"]
        assert cert.Certificate == VPN_CERTIFICATE_API_RESPONSE["Certificate"]
        assert cert.ExpirationTime == VPN_CERTIFICATE_API_RESPONSE["ExpirationTime"]
        assert cert.RefreshTime == VPN_CERTIFICATE_API_RESPONSE["RefreshTime"]
        assert cert.Mode == VPN_CERTIFICATE_API_RESPONSE["Mode"]
        assert cert.DeviceName == VPN_CERTIFICATE_API_RESPONSE["DeviceName"]
        assert cert.ServerPublicKeyMode == VPN_CERTIFICATE_API_RESPONSE["ServerPublicKeyMode"]
        assert cert.ServerPublicKey == VPN_CERTIFICATE_API_RESPONSE["ServerPublicKey"]

    def test_vpncertificate_to_dict(self):
        assert VPNCertificate.from_dict(VPN_CERTIFICATE_API_RESPONSE).to_dict() == VPN_CERTIFICATE_API_RESPONSE

    def test_secrets_from_dict(self):
        secrets = VPNSecrets.from_dict(VPN_SECRETS_DICT)
        assert secrets.ed25519_privatekey == "rNW3dL5A3dUrQX3ZKbVAFLjSFJdvDU5JzjrRrnI+cos="

    def test_secrets_to_dict(self):
        assert VPNSecrets.from_dict(VPN_SECRETS_DICT).to_dict() == VPN_SECRETS_DICT

    def test_sessions_from_dict(self):
        sessions = VPNSessions.from_dict(VPN_SESSIONS_API_RESPONSE)
        assert(len(sessions.Sessions)==2)
        assert(sessions.Sessions[0].ExitIP=='1.2.3.4')
        assert(sessions.Sessions[1].ExitIP=='5.6.7.8')

    def test_location_from_dict(self):
        location = VPNLocation.from_dict(VPN_LOCATION_API_RESPONSE)
        assert location.IP == VPN_LOCATION_API_RESPONSE["IP"]
        assert location.Country == VPN_LOCATION_API_RESPONSE["Country"]
        assert location.ISP == VPN_LOCATION_API_RESPONSE["ISP"]

    def test_location_to_dict(self):
        assert VPNLocation.from_dict(VPN_LOCATION_API_RESPONSE).to_dict() == VPN_LOCATION_API_RESPONSE


class TestVpnAccount:
    def test_vpn_session___setstate__(self):
        vpnsession = VPNSession()
        vpndata={
            "vpn" : {
                "vpninfo": VPN_API_RESPONSE,
                "certificate": VPN_CERTIFICATE_API_RESPONSE,
                "location": VPN_LOCATION_API_RESPONSE,
                "secrets": VPN_SECRETS_DICT
            }
        }
        vpnsession.__setstate__(vpndata)

        vpn_account = vpnsession.vpn_account
        assert vpn_account.max_tier == 0
        assert vpn_account.max_connections == 2
        assert vpn_account.plan_name == vpndata["vpn"]["vpninfo"]["VPN"]["PlanName"]
        assert vpn_account.plan_title == vpndata["vpn"]["vpninfo"]["VPN"]["PlanTitle"]
        assert not vpn_account.delinquent
        assert vpn_account.location.to_dict() == vpndata["vpn"]["location"]
        vpncredentials = vpnsession.vpn_account.vpn_credentials
        assert vpncredentials.userpass_credentials.username == vpndata["vpn"]["vpninfo"]["VPN"]["Name"]
        assert vpncredentials.userpass_credentials.password == vpndata["vpn"]["vpninfo"]["VPN"]["Password"]
        assert vpncredentials.pubkey_credentials.ed_255519_private_key == vpndata["vpn"]["secrets"]["ed25519_privatekey"]


class TestPubkeyCredentials:
    def test_certificate_fingerprint_mismatch(self):
        # Generate a new keypair. This means its fingerprint won't match the one
        # from /vpn/v1/certificate.
        with pytest.raises(VPNCertificateFingerprintError):
            VPNPubkeyCredentials(
                api_certificate=VPNCertificate.from_dict(VPN_CERTIFICATE_API_RESPONSE),
                # A new keypair is generated: its fingerprint won't match the one returned by /vpn/v1/certificate.
                secrets=VPNSecrets(),
            )

    def test_certificate_duration(self):
        pubkey_credentials = VPNPubkeyCredentials(
            api_certificate=VPNCertificate.from_dict(VPN_CERTIFICATE_API_RESPONSE),
            # A new keypair is generated: its fingerprint won't match the one returned by /vpn/v1/certificate.
            secrets=VPNSecrets.from_dict(VPN_SECRETS_DICT),
        )

        assert(pubkey_credentials.certificate_duration == 86401.0)



"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from datetime import datetime, timezone, timedelta
from unittest.mock import Mock, patch

import pytest

from proton.vpn.session.notifications_fetcher import (
    NotificationsFetcher, Notifications, NpsSurvey, NotificationType
)
from proton.vpn.session.exceptions import NotificationDecodeError

NOTIFICATION_ID = "NpsSurvey"
_now = datetime.now(tz=timezone.utc)
START_TIME_PAST = int((_now - timedelta(days=1)).timestamp())
END_TIME_FUTURE = int((_now + timedelta(days=1)).timestamp())


@pytest.fixture
def notification_dict():
    return {
        "NotificationID": NOTIFICATION_ID,
        "Type": NotificationType.NPSSURVEY,
        "StartTime": START_TIME_PAST,
        "EndTime": END_TIME_FUTURE,
    }


@pytest.fixture
def api_data(notification_dict):
    return {
        "Code": 1000,
        "Notifications": [notification_dict],
    }


@pytest.fixture
def api_data_with_seen(notification_dict):
    return {
        "Code": 1000,
        "Notifications": [{**notification_dict, "seen": True}],
    }


# ---------------------------------------------------------------------------
# NpsSurvey
# ---------------------------------------------------------------------------

def test_nps_survey_from_dict_converts_unix_timestamps_to_aware_datetimes(notification_dict):
    survey = NpsSurvey.from_dict(notification_dict)

    assert survey.start_time.tzinfo == timezone.utc
    assert survey.end_time.tzinfo == timezone.utc


def test_nps_survey_is_active_within_time_window():
    now = datetime.now(tz=timezone.utc)
    survey = NpsSurvey(
        survey_id="x",
        start_time=now - timedelta(hours=1),
        end_time=now + timedelta(hours=1),
        seen=False,
    )
    assert survey.is_active is True


def test_nps_survey_is_not_active_before_start_time():
    now = datetime.now(tz=timezone.utc)
    survey = NpsSurvey(
        survey_id="x",
        start_time=now + timedelta(hours=1),
        end_time=now + timedelta(hours=2),
        seen=False,
    )
    assert survey.is_active is False


def test_nps_survey_is_not_active_after_end_time():
    now = datetime.now(tz=timezone.utc)
    survey = NpsSurvey(
        survey_id="x",
        start_time=now - timedelta(hours=2),
        end_time=now - timedelta(hours=1),
        seen=False,
    )
    assert survey.is_active is False


def test_nps_survey_seen_defaults_to_false_when_not_in_dict(notification_dict):
    survey = NpsSurvey.from_dict(notification_dict)
    assert survey.seen is False


def test_nps_survey_from_dict_raises_on_missing_required_key():
    with pytest.raises(NotificationDecodeError):
        NpsSurvey.from_dict(
            {
                "StartTime": START_TIME_PAST,
                "EndTime": END_TIME_FUTURE
            }
        )


def test_nps_survey_from_dict_raises_on_invalid_timestamp():
    with pytest.raises(NotificationDecodeError):
        NpsSurvey.from_dict(
            {
                "NotificationID": NOTIFICATION_ID,
                "StartTime": "not-a-timestamp",
                "EndTime": END_TIME_FUTURE
            }
        )


# ---------------------------------------------------------------------------
# NotificationsFetcher — cache
# ---------------------------------------------------------------------------


@patch("proton.vpn.session.notifications_fetcher.rest_api_request")
@pytest.mark.asyncio
async def test_fetch_saves_response_to_cache(mock_rest_api_request, api_data):
    mock_cache_handler = Mock()
    mock_refresh_calculator = Mock()
    mock_refresh_calculator.get_expiration_time.return_value = 10

    mock_rest_api_request.return_value = api_data

    fetcher = NotificationsFetcher(Mock(), mock_refresh_calculator, mock_cache_handler)
    await fetcher.fetch()

    mock_cache_handler.save.assert_called_once()


def test_load_from_cache_returns_notifications_from_cache(api_data):
    mock_cache_handler = Mock()
    mock_cache_handler.load.return_value = api_data

    fetcher = NotificationsFetcher(Mock(), Mock(), mock_cache_handler)
    notifications = fetcher.load_from_cache()

    assert notifications.data["Notifications"] == api_data["Notifications"]


def test_load_from_cache_returns_empty_expired_notifications_when_no_cache_found():
    mock_cache_handler = Mock()
    mock_cache_handler.load.return_value = None

    fetcher = NotificationsFetcher(Mock(), Mock(), mock_cache_handler)
    notifications = fetcher.load_from_cache()

    assert notifications.get_nps_survey_notifications() == []
    assert notifications.is_expired


# ---------------------------------------------------------------------------
# NotificationsFetcher — seen state
# ---------------------------------------------------------------------------

def test_set_notification_seen_marks_notification_as_seen(api_data):
    mock_cache_handler = Mock()
    mock_cache_handler.load.return_value = api_data

    fetcher = NotificationsFetcher(Mock(), Mock(), mock_cache_handler)
    notifications = fetcher.load_from_cache()
    fetcher.set_notification_seen(NOTIFICATION_ID)

    assert notifications.data["Notifications"][0]["seen"] is True


def test_set_notification_seen_persists_seen_state_to_cache(api_data):
    mock_cache_handler = Mock()
    mock_cache_handler.load.return_value = api_data

    fetcher = NotificationsFetcher(Mock(), Mock(), mock_cache_handler)
    fetcher.load_from_cache()

    mock_cache_handler.save.reset_mock()
    fetcher.set_notification_seen(NOTIFICATION_ID)

    mock_cache_handler.save.assert_called_once()
    saved_data = mock_cache_handler.save.call_args[0][0]
    assert saved_data["Notifications"][0]["seen"] is True


@patch("proton.vpn.session.notifications_fetcher.rest_api_request")
@pytest.mark.asyncio
async def test_fetch_propagates_seen_state_to_new_api_snapshot(mock_rest_api_request, api_data):
    mock_cache_handler = Mock()
    mock_refresh_calculator = Mock()
    mock_refresh_calculator.get_expiration_time.return_value = 10

    mock_cache_handler.load.return_value = api_data

    fresh_api_data = api_data.copy()

    fetcher = NotificationsFetcher(Mock(), mock_refresh_calculator, mock_cache_handler)
    fetcher.load_from_cache()
    fetcher.set_notification_seen(NOTIFICATION_ID)

    # API returns fresh snapshot for the same notification ID, without seen
    mock_rest_api_request.return_value = fresh_api_data

    notifications = await fetcher.fetch()

    assert notifications.data["Notifications"][0]["seen"] is True


@patch("proton.vpn.session.notifications_fetcher.rest_api_request")
@pytest.mark.asyncio
async def test_fetch_does_not_mark_unseen_notification_as_seen(mock_rest_api_request, api_data):
    mock_cache_handler = Mock()
    mock_refresh_calculator = Mock()
    mock_refresh_calculator.get_expiration_time.return_value = 10

    mock_cache_handler.load.return_value = api_data

    fetcher = NotificationsFetcher(Mock(), mock_refresh_calculator, mock_cache_handler)
    fetcher.load_from_cache()
    # Deliberately do NOT call set_notification_seen

    mock_rest_api_request.return_value = api_data

    notifications = await fetcher.fetch()

    assert notifications.data["Notifications"][0].get("seen", False) is False


@patch("proton.vpn.session.notifications_fetcher.rest_api_request")
@pytest.mark.asyncio
async def test_fetch_does_not_propagate_seen_to_different_notification_id(mock_rest_api_request, api_data):
    mock_cache_handler = Mock()
    mock_refresh_calculator = Mock()
    mock_refresh_calculator.get_expiration_time.return_value = 10

    mock_cache_handler.load.return_value = api_data

    fetcher = NotificationsFetcher(Mock(), mock_refresh_calculator, mock_cache_handler)
    fetcher.load_from_cache()
    fetcher.set_notification_seen(NOTIFICATION_ID)

    # API returns a different notification ID
    different_id_data = {
        "Code": 1000,
        "Notifications": [
            {
                "NotificationID": "survey-2",
                "Type": NotificationType.NPSSURVEY,
                "StartTime": START_TIME_PAST,
                "EndTime": END_TIME_FUTURE,
            }
        ],
    }
    mock_rest_api_request.return_value = different_id_data

    notifications = await fetcher.fetch()

    assert notifications.data["Notifications"][0].get("seen", False) is False


# ---------------------------------------------------------------------------
# Notifications.get_nps_survey_notifications
# ---------------------------------------------------------------------------

def test_get_nps_survey_notifications_returns_only_nps_type():
    data = {
        "Notifications": [
            {
                "NotificationID": NOTIFICATION_ID,
                "Type": NotificationType.NPSSURVEY,
                "StartTime": START_TIME_PAST,
                "EndTime": END_TIME_FUTURE,
            },
            {
                "NotificationID": "other-1",
                "Type": 99,
                "StartTime": START_TIME_PAST,
                "EndTime": END_TIME_FUTURE,
            },
        ]
    }

    notifications = Notifications(data)
    results = notifications.get_nps_survey_notifications()

    assert len(results) == 1
    assert results[0].survey_id == NOTIFICATION_ID


def test_get_nps_survey_notifications_skips_notification_with_missing_type():
    data = {
        "Notifications": [
            {
                "NotificationID": NOTIFICATION_ID,
                "StartTime": START_TIME_PAST,
                "EndTime": END_TIME_FUTURE,
            }
        ]
    }

    notifications = Notifications(data)
    assert notifications.get_nps_survey_notifications() == []


def test_get_nps_survey_notifications_skips_notification_with_invalid_type():
    data = {
        "Notifications": [
            {
                "NotificationID": NOTIFICATION_ID,
                "Type": NotificationType.INVALID,
                "StartTime": START_TIME_PAST,
                "EndTime": END_TIME_FUTURE,
            }
        ]
    }

    notifications = Notifications(data)
    assert notifications.get_nps_survey_notifications() == []


def test_get_nps_survey_notifications_skips_malformed_notifications():
    data = {
        "Notifications": [
            {
                # Missing NotificationID — should be skipped, not raise
                "Type": NotificationType.NPSSURVEY,
                "StartTime": START_TIME_PAST,
                "EndTime": END_TIME_FUTURE,
            }
        ]
    }

    notifications = Notifications(data)
    assert notifications.get_nps_survey_notifications() == []


def test_get_nps_survey_notifications_reflects_seen_state():
    data = {
        "Notifications": [
            {
                "NotificationID": NOTIFICATION_ID,
                "Type": NotificationType.NPSSURVEY,
                "StartTime": START_TIME_PAST,
                "EndTime": END_TIME_FUTURE,
                "seen": True,
            }
        ]
    }

    notifications = Notifications(data)
    results = notifications.get_nps_survey_notifications()

    assert results[0].seen is True



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock

from proton.vpn.connection.interfaces import (Settings, VPNCredentials,
                                              VPNPubkeyCredentials, VPNServer,
                                              VPNUserPassCredentials, Features)
import pathlib
import os
from collections import namedtuple

CWD = str(pathlib.Path(__file__).parent.absolute())
PERSISTANCE_CWD = os.path.join(
    CWD,
    "connection_persistence"
)
OpenVPNPorts = namedtuple("OpenVPNPorts", "udp tcp")
WireGuardPorts = namedtuple("WireGuardPorts", "udp tcp")


class MalformedVPNCredentials:
    pass


class MalformedVPNServer:
    pass





class MockVPNPubkeyCredentials(VPNPubkeyCredentials):
    @property
    def certificate_pem(self):
        return "pem-cert"

    @property
    def wg_private_key(self):
        return "wg-private-key"

    @property
    def openvpn_private_key(self):
        return "ovpn-private-key"

    def get_ed25519_sk_pem(self, password=None):
        return "encrypted-ovpn-private-key"


class MockVPNUserPassCredentials(VPNUserPassCredentials):
    @property
    def username(self):
        return "test-username"

    @property
    def password(self):
        return "test-password"


class MockVpnCredentials(VPNCredentials):
    @property
    def pubkey_credentials(self):
        return MockVPNPubkeyCredentials()

    @property
    def userpass_credentials(self):
        return MockVPNUserPassCredentials()


class MockSettings(Settings):
    @property
    def dns_custom_ips(self):
        return ["1.1.1.1", "10.10.10.10"]

    @property
    def features(self):
        return Mock()



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock

from proton.vpn.connection import events
from proton.vpn.connection.enum import StateMachineEventEnum
import pytest

from proton.vpn.connection.events import EventContext

context = EventContext(connection=Mock())


def test_base_class_missing_event():
    class DummyEvent(events.Event):
        pass

    with pytest.raises(AttributeError):
        DummyEvent(context)


def test_base_class_expected_event():
    custom_event = "test_event"

    class DummyEvent(events.Event):
        type = custom_event

    assert DummyEvent(context).type == custom_event


@pytest.mark.parametrize(
    "event_class, expected_event",
    [
        (events.Up.type, StateMachineEventEnum.UP),
        (events.Down.type, StateMachineEventEnum.DOWN),
        (events.Connected.type, StateMachineEventEnum.CONNECTED),
        (events.Disconnected.type, StateMachineEventEnum.DISCONNECTED),
        (events.Timeout.type, StateMachineEventEnum.TIMEOUT),
        (events.AuthDenied.type, StateMachineEventEnum.AUTH_DENIED),
        (events.UnexpectedError.type, StateMachineEventEnum.UNEXPECTED_ERROR),
    ]
)
def test_individual_events(event_class, expected_event):
    assert event_class == expected_event



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import pytest
import tempfile
from proton.vpn import logging
import logging as _logging


@pytest.fixture(scope="module")
def test_logger():
    with tempfile.TemporaryDirectory() as tmpdir:
        logging.config("test-file", logdirpath=tmpdir)
        logger = logging.getLogger(__name__)
        logger.setLevel(_logging.DEBUG)
        yield logger


def log_debug(logger):
    logger.debug("test-message-debug", category="CAT", event="EV")


def log_info(logger):
    logger.info("test-message-info", category="CAT", event="EV")


def log_warning(logger):
    logger.warning("warning", category="CAT", event="EV")


def log_error(logger):
    logger.error("error", category="CAT", event="EV")


def log_critical(logger):
    logger.critical("critical", category="CAT", event="EV")


def log_exception(logger):
    try:
        raise Exception("test")
    except Exception:
        logger.exception("exception", category="CAT", event="EV")


def test_debug_with_custom_properties(caplog, test_logger):
    caplog.clear()
    log_debug(test_logger)
    for record in caplog.records:
        assert record.levelname == "DEBUG"
    assert len(caplog.records) == 1


def test_info_with_custom_properties(caplog, test_logger):
    caplog.clear()
    log_info(test_logger)
    for record in caplog.records:
        assert record.levelname == "INFO"
    assert len(caplog.records) == 1


def test_warning_with_custom_properties(caplog, test_logger):
    caplog.clear()
    log_warning(test_logger)
    for record in caplog.records:
        assert record.levelname == "WARNING"
    assert len(caplog.records) == 1


def test_error_with_custom_properties(caplog, test_logger):
    caplog.clear()
    log_error(test_logger)
    for record in caplog.records:
        assert record.levelname == "ERROR"
    assert len(caplog.records) == 1


def test_critical_with_custom_properties(caplog, test_logger):
    caplog.clear()
    log_critical(test_logger)
    for record in caplog.records:
        assert record.levelname == "CRITICAL"
    assert len(caplog.records) == 1


def test_exception_with_custom_properties(caplog, test_logger):
    caplog.clear()
    log_exception(test_logger)
    for record in caplog.records:
        assert record.levelname == "ERROR"
    assert len(caplog.records) == 1
    assert "exception" in caplog.text


def test_debug_with_only_message_logging_properties(caplog, test_logger):
    caplog.clear()
    test_logger.debug(msg="test-default-debug")
    for record in caplog.records:
        assert record.levelname == "DEBUG"
    assert len(caplog.records) == 1
    assert "test-default-debug" in caplog.text


def test_debug_with_no_logging_properties(caplog, test_logger):
    caplog.clear()
    test_logger.debug(msg="")
    assert len(caplog.records) == 1
    assert "" in caplog.text



#!/usr/bin/env python3
'''
This program generates a deb changelog file, and rpm spec file and a
CHANGELOG.md file for this project.

It reads versions.yml.
'''
import os
import yaml
import devtools.versions as versions

# The root of this repo
ROOT = os.path.dirname(
    os.path.dirname(os.path.realpath(__file__))
)

NAME      = "proton-vpn-api-core"  # Name of this application.
VERSIONS  = os.path.join(ROOT, "versions.yml")  # Name of this applications versions.yml
RPM       = os.path.join(ROOT, "rpmbuild", "SPECS", "package.spec")  # Path of spec file for rpm.
RPM_TMPLT = os.path.join(ROOT, "rpmbuild", "SPECS", "package.spec.template")  # Path of template spec file for rpm.
DEB       = os.path.join(ROOT, "debian", "changelog")  # Path of debian changelog.
MARKDOWN  = os.path.join(ROOT, "CHANGELOG.md",)  # Path of CHANGELOG.md.

def build():
    '''
    This is what generates the rpm spec, deb changelog and
    markdown CHANGELOG.md file.
    '''
    with open(VERSIONS, encoding="utf-8") as versions_file:

        # Load versions.yml
        versions_yml = list(yaml.safe_load_all(versions_file))

        # Validate the versions.yml file
        #
        # This is a lint of the versions.yml and catches errors
        # that might not be found in the changelog generation process
        versions.validate_versions(versions_yml)

        # Make our files
        versions.build_rpm(RPM,      versions_yml, RPM_TMPLT)
        versions.build_deb(DEB,      versions_yml, NAME)
        versions.build_mkd(MARKDOWN, versions_yml)


if __name__ == "__main__":
    build()



import os

import pytest
from proton.vpn.backend.networkmanager.protocol.wireguard.wireguard import FWMARK_ENV_VAR, get_fwmark_from_env_var


def test_get_fwmark_from_env_var_returns_env_var_value_if_available_and_valid():
    os.environ[FWMARK_ENV_VAR] = "51821"
    assert get_fwmark_from_env_var() == 51821


@pytest.mark.parametrize(
        "env_var_value", [
            "51820",  # too small
            str(2**32),  # too big
            "#!!?"  # invalid int,
        ]
)
def test_get_fwmark_from_env_var_returns_None_if_env_var_contains_invalid_value(env_var_value):
    os.environ[FWMARK_ENV_VAR] = env_var_value
    assert get_fwmark_from_env_var() is None


"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from proton.vpn.connection import VPNServer, ProtocolPorts
from proton.vpn.connection.persistence import ConnectionPersistence, ConnectionParameters


@pytest.fixture
def temp_dir() -> str:
    with TemporaryDirectory(suffix=__name__) as temp_dir:
        yield f"{temp_dir}"


def test_load(temp_dir: str):
    with open(os.path.join(temp_dir, ConnectionPersistence.FILENAME), "w") as f:
        f.write('''{
  "connection_id": "connection_id",
  "backend": "backend",
  "protocol": "protocol",
  "server": {
    "server_ip": "1.2.3.4",
    "openvpn_ports": {
      "udp": [12345],
      "tcp": [80]
    },
    "wireguard_ports": {
      "udp": [54321],
      "tcp": [81]
    },
    "domain": "server.domain",
    "x25519pk": "public_key",
    "server_id": "server_id",
    "server_name": "server_name",
    "has_ipv6_support": "0",
    "label": "label"
  }
}''')

    connection_persistence = ConnectionPersistence(persistence_directory=temp_dir)
    persisted_parameters = connection_persistence.load()

    assert persisted_parameters.connection_id == "connection_id"
    assert persisted_parameters.backend == "backend"
    assert persisted_parameters.protocol == "protocol"
    assert persisted_parameters.server.server_ip == "1.2.3.4"
    assert persisted_parameters.server.openvpn_ports.udp == [12345]
    assert persisted_parameters.server.openvpn_ports.tcp == [80]
    assert persisted_parameters.server.wireguard_ports.udp == [54321]
    assert persisted_parameters.server.wireguard_ports.tcp == [81]
    assert persisted_parameters.server.domain == "server.domain"
    assert persisted_parameters.server.x25519pk == "public_key"
    assert persisted_parameters.server.server_id == "server_id"
    assert persisted_parameters.server.server_name == "server_name"
    assert persisted_parameters.server.label == "label"


def test_load_returns_none_and_logs_error_when_persistence_file_contains_invalid_json(temp_dir, caplog):
    with open(os.path.join(temp_dir, ConnectionPersistence.FILENAME), "w") as f:
        f.write('{"conn')

    connection_persistence = ConnectionPersistence(persistence_directory=temp_dir)
    persisted_parameters = connection_persistence.load()

    assert not persisted_parameters
    assert len([r for r in caplog.records if r.levelname == "WARNING"]) == 1


def test_load_returns_none_and_logs_error_when_persistence_file_misses_expected_parameters(temp_dir):
    with open(os.path.join(temp_dir, ConnectionPersistence.FILENAME), "w") as f:
        f.write('{"foo": "bar"}')

    connection_persistence = ConnectionPersistence(persistence_directory=temp_dir)
    persisted_parameters = connection_persistence.load()

    assert not persisted_parameters


def test_save_(temp_dir: str):
    connection_parameters = ConnectionParameters(
        connection_id="connection_id",
        backend="backend",
        protocol="protocol",
        server=VPNServer(
            server_ip="1.2.3.4",
            openvpn_ports=ProtocolPorts(
                udp=[12345],
                tcp=[80],
                tls=[8080],
            ),
            wireguard_ports=ProtocolPorts(
                udp=[54321],
                tcp=[81],
                tls=[8080],
            ),
            domain="server.domain",
            x25519pk="public_key",
            server_id="server_id",
            server_name="server_name",
            has_ipv6_support=False,
            label="label"
        )
    )

    connection_persistence = ConnectionPersistence(persistence_directory=temp_dir)
    connection_persistence.save(connection_parameters)

    with open(os.path.join(temp_dir, ConnectionPersistence.FILENAME)) as f:
        persistence_file_content = json.load(f)

        assert connection_parameters.connection_id == persistence_file_content["connection_id"]
        assert connection_parameters.backend == persistence_file_content["backend"]
        assert connection_parameters.protocol == persistence_file_content["protocol"]
        assert connection_parameters.server.server_ip == persistence_file_content["server"]["server_ip"]
        assert connection_parameters.server.openvpn_ports.udp == persistence_file_content["server"]["openvpn_ports"]["udp"]
        assert connection_parameters.server.openvpn_ports.tcp == persistence_file_content["server"]["openvpn_ports"]["tcp"]
        assert connection_parameters.server.wireguard_ports.udp == persistence_file_content["server"]["wireguard_ports"]["udp"]
        assert connection_parameters.server.wireguard_ports.tcp == persistence_file_content["server"]["wireguard_ports"]["tcp"]
        assert connection_parameters.server.domain == persistence_file_content["server"]["domain"]
        assert connection_parameters.server.x25519pk == persistence_file_content["server"]["x25519pk"]
        assert connection_parameters.server.server_id == persistence_file_content["server"]["server_id"]
        assert connection_parameters.server.server_name == persistence_file_content["server"]["server_name"]
        assert connection_parameters.server.label == persistence_file_content["server"]["label"]


def test_remove(temp_dir: str):
    persistence_file_path = Path(temp_dir) / ConnectionPersistence.FILENAME
    persistence_file_path.touch()
    assert persistence_file_path.is_file()

    connection_persistence = ConnectionPersistence(persistence_directory=temp_dir)
    connection_persistence.remove()

    assert not persistence_file_path.exists()


def test_remove_logs_a_warning_when_persistence_file_was_not_found(
        temp_dir:str, caplog
):
    connection_persistence = ConnectionPersistence(persistence_directory=temp_dir)
    connection_persistence.remove()

    assert len(caplog.records) == 1
    assert len([r for r in caplog.records if r.levelname == "WARNING"]) == 1



"""
Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from dataclasses import asdict
import pytest
from proton.vpn.session.dataclasses import VPNSettings, VPNInfo


@pytest.fixture
def vpninfo_data():
    return {
        "ExpirationTime": 1,
        "Name": "random_user",
        "Password": "asdKJkjb12",
        "GroupID": "test-group",
        "Status": 1,
        "PlanName": "test plan",
        "PlanTitle": "test title",
        "MaxTier": 1,
        "MaxConnect": 1,
        "Groups": ["group1", "group2"],
        "NeedConnectionAllocation": False,
    }


def test_vpninfo_deserializes_expected_dict_keys(vpninfo_data):
    vpninfo = VPNInfo.from_dict(vpninfo_data)

    assert asdict(vpninfo) == vpninfo_data


def test_vpninfo_deserialize_should_not_crash_with_unexpected_dict_keys(vpninfo_data):
    vpninfo_data["unexpected_keyword"] = "keyword and data"
    VPNInfo.from_dict(vpninfo_data)


@pytest.fixture
def vpnsettings_data(vpninfo_data):
    return {
        "VPN": vpninfo_data,
        "Services": 1,
        "Subscribed": 1,
        "Delinquent": 0,
        "HasPaymentMethod": 1,
        "Credit": 1234,
        "Currency": "€",
        "Warnings": [],
    }


def test_vpnsettings_deserializes_expected_dict_keys(vpnsettings_data):
    vpnsettings = VPNSettings.from_dict(vpnsettings_data)

    assert asdict(vpnsettings) == vpnsettings_data


def test_vpnsettings_deserialize_should_not_crash_with_unexpected_dict_keys(vpnsettings_data):
    vpnsettings_data["unexpected_keyword"] = "keyword and data"
    VPNSettings.from_dict(vpnsettings_data)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import pytest
from unittest.mock import Mock
from proton.vpn.core.registry import Registry

# pylint: disable=C0116


class _Interface:  # pylint: disable=R0903
    pass


class _LowPriorityEntry(_Interface):  # pylint: disable=R0903
    @classmethod
    def get_key(cls):
        return "low"

    @classmethod
    def get_priority(cls):
        return 10

    @classmethod
    def validate(cls):
        return True


class _HighPriorityEntry(_Interface):  # pylint: disable=R0903
    @classmethod
    def get_key(cls):
        return "high"

    @classmethod
    def get_priority(cls):
        return 1

    @classmethod
    def validate(cls):
        return True


class _InvalidEntry(_Interface):  # pylint: disable=R0903
    @classmethod
    def get_key(cls):
        return "invalid"

    @classmethod
    def get_priority(cls):
        return 1

    @classmethod
    def validate(cls):
        return False


class _NoPriorityEntry(_Interface):  # pylint: disable=R0903
    @classmethod
    def get_key(cls):
        return "no_priority"

    @classmethod
    def get_priority(cls):
        return None

    @classmethod
    def validate(cls):
        return True


class _UnrelatedEntry:  # pylint: disable=R0903
    @classmethod
    def get_key(cls):
        return "unrelated"

    @classmethod
    def get_priority(cls):
        return 1

    @classmethod
    def validate(cls):
        return True


def test_register_stores_entry_under_given_key():
    registry = Registry()
    registry.register(_LowPriorityEntry)
    assert registry.get("low") is _LowPriorityEntry


def test_iter_returns_entries_ordered_by_priority():
    registry = Registry()
    registry.register(_LowPriorityEntry)
    registry.register(_HighPriorityEntry)

    result = list(registry.iter(_Interface))

    assert result == [_HighPriorityEntry, _LowPriorityEntry]


def test_iter_skips_entries_that_fail_validation():
    registry = Registry()
    registry.register(_InvalidEntry)
    registry.register(_HighPriorityEntry)

    result = list(registry.iter(_Interface))

    assert result == [_HighPriorityEntry]


def test_iter_skips_entries_not_subclass_of_interface():
    registry = Registry()
    registry.register(_UnrelatedEntry)
    registry.register(_HighPriorityEntry)

    result = list(registry.iter(_Interface))

    assert result == [_HighPriorityEntry]


def test_iter_skips_entries_with_none_priority():
    registry = Registry()
    registry.register(_NoPriorityEntry)
    registry.register(_HighPriorityEntry)

    result = list(registry.iter(_Interface))

    assert result == [_HighPriorityEntry]


def test_get_returns_entry_matching_key():
    registry = Registry()
    registry.register(_LowPriorityEntry)
    registry.register(_HighPriorityEntry)

    result = registry.get("low")

    assert result is _LowPriorityEntry


def test_get_raises_runtime_error_when_key_not_found():
    registry = Registry()
    registry.register(_LowPriorityEntry)

    with pytest.raises(RuntimeError):
        registry.get("missing")


def test_register_from_registrars_calls_each_registrar_with_registry():
    registry = Registry()
    registrar_a = Mock()
    registrar_b = Mock()

    registry.register_from_registrars([registrar_a, registrar_b])

    registrar_a.assert_called_once_with(registry)
    registrar_b.assert_called_once_with(registry)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import os
import tempfile

import pytest
from proton.vpn.core.cache_handler import CacheHandler


class TestCacheHandler:
    @pytest.fixture
    def dir_path(self):
        configfolder = tempfile.TemporaryDirectory(prefix="test_cache_handler")
        yield configfolder
        configfolder.cleanup()

    @pytest.fixture
    def cache_filepath(self, dir_path):
        return os.path.join(dir_path.name, "test_cache_file.json")

    def test_save_new_cache(self, cache_filepath):
        cache_handler = CacheHandler(cache_filepath)
        cache_handler.save({"save_cache": "dummy-data"})
        with open(cache_filepath, "r") as f:
            content = json.load(f)
            assert "save_cache" in content
            assert "dummy-data" == content["save_cache"]

    def test_load_stored_cache(self, cache_filepath):
        cache_handler = CacheHandler(cache_filepath)
        with open(cache_filepath, "w") as f:
            json.dump({"load_cache": "dummy-data"}, f)

        data = cache_handler.load()
        assert "load_cache" in data
        assert "dummy-data" == data["load_cache"]

    def test_load_cache_with_missing_file(self, cache_filepath):
        cache_handler = CacheHandler(cache_filepath)
        assert not cache_handler.load()

    def test_remove_cache(self, cache_filepath):
        cache_handler = CacheHandler(cache_filepath)
        with open(cache_filepath, "w") as f:
            json.dump({"load_cache": "dummy-data"}, f)

        cache_handler.remove()

        assert not os.path.isfile(cache_filepath)



"""
Registry for modular components of the VPN API.

Any type can be registered as long as it implements the `Entry` protocol, the
type can then be retrived by its key or by iterating over the available
implementations.

This is a simpler and more dynamic alternative to using entry points.

Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import importlib

from abc import abstractmethod
from typing import Iterator, Optional, Protocol


class Entry(Protocol):  # pylint: disable=R0903
    """
    Registry entries must implement this protocol.
    """
    @classmethod
    @abstractmethod
    def get_key(cls) -> str:
        """
        Type name the implementation will be registered under.

        To be implemented by subclasses.

        :return: The type name the implementation will be registered under.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def get_priority(cls) -> int:
        """
        Priority of the implementation.

        To be implemented by subclasses.

        :return: The priority of the implementation.
            Lower values indicate higher priority.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def validate(cls) -> bool:
        """
        Determines whether the implementation is valid or not.
        To be implemented by subclasses.

        If this method returns `False` then the implementation
        will be skipped.

        :return: `True` if the implementation is valid or `False` otherwise.
        """
        raise NotImplementedError


class Registry:
    """
    Registry of implementations.

    This class is used to keep track of the available implementations.
    """

    def __init__(self):
        self._registry = {}

    def register(self, entry_type: type[Entry]):
        """Registers an implementation."""

        self._registry[entry_type.get_key()] = entry_type  # pylint: disable=W0212

    def register_from_registrars(self, registrars):
        """Registers implementations by calling each registrar with this registry."""
        for registrar in registrars:
            registrar(self)

    def register_from_module(self, module: str):
        """
        Registers implementations under the given root module.
        """
        self.register_from_registrars(
            # pylint: disable-next=line-too-long
            [importlib.import_module(module).register]  # noqa: E501 # nosemgrep: python.lang.security.audit.non-literal-import.non-literal-import
        )

    def get(self, key: str) -> Optional[type[Entry]]:
        """
        Returns the implementation for the given key.

        If there's no valid implementation for the given key, it returns None.
        """
        if entry := self._registry.get(key, None):
            if entry.validate():
                return entry

        raise RuntimeError(f"No valid implementation found for {key}")

    def iter(self, interface: type) -> Iterator[type[Entry]]:
        """
        Returns an iterator over the available entries ordered by
        priority and filtered by validity.
        """
        def validate_and_priority(cls):
            return (
                issubclass(cls, interface)
                and cls.validate()  # pylint: disable=W0212
                and cls.get_priority() is not None  # pylint: disable=W0212
            )

        def priority(cls):
            return cls.get_priority()  # pylint: disable=W0212

        return iter(
            sorted(
                filter(
                    validate_and_priority,
                    self._registry.values(),
                ),
                key=priority
            )
        )



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Type
from unittest.mock import Mock, call, AsyncMock

import pytest

from proton.vpn.connection import states, events
from proton.vpn.connection.enum import KillSwitchSetting
from proton.vpn.connection.exceptions import ConcurrentConnectionsError
from proton.vpn.core.settings.split_tunneling import SplitTunneling as SplitTunnelingSetting
from proton.vpn.split_tunneling.exceptions import SplitTunnelingError


def test_state_subclass_raises_exception_when_missing_state():
    class DummyState(states.State):
        pass

    with pytest.raises(TypeError):
        DummyState(states.StateContext())


def test_state_on_event_logs_warning_when_event_did_not_cause_state_transition(caplog):
    class DummyState(states.State):
        type = Mock()

        def _on_event(self, event: events.Event) -> states.State:
            return self

    state = DummyState(states.StateContext())

    new_state = state.on_event(events.Up(events.EventContext(connection=Mock())))

    assert new_state is state
    warnings = [record for record in caplog.records if record.levelname == "WARNING"]
    assert len(warnings) == 1
    assert "state received unexpected event" in warnings[0].message


@pytest.mark.parametrize(
    "event_type, concurrent_connections_error_expected", [
        (event_type, event_type != events.Up) for event_type in events.EVENT_TYPES
    ]
)
def test_state_on_event_raises_concurrent_connections_error_when_multiple_connections_are_detected(
        event_type, concurrent_connections_error_expected
):
    """
    All state instance raise an exception if they receive an event carrying a connection that's not
    the same as the one the state instance already has on its context. The reason for this is that
    the current state should be receiving state updates from the same connection that led to this
    state.

    The exception to this rule is the Up event, since the goal of the Up event is to start a new
    connection.
    """
    # In this case, the concrete state instance doesn't matter, since this check is done in
    # the base State class.
    state = states.Connected(states.StateContext(connection=Mock()))
    event = event_type(events.EventContext(connection=Mock()))

    try:
        state.on_event(event)
        error_raised = False
    except ConcurrentConnectionsError:
        error_raised = True

    assert error_raised is concurrent_connections_error_expected


def assert_state_transition(
        state_type: Type[states.State],
        event_type: Type[events.Event],
        expected_next_state_type: Type[states.State]
):
    """Asserts that when calling the `on_event` method on an instance of `state_type` passing it
    an instance of `event_type` then the result is an instance of `expected_next_state_type`."""
    connection = Mock()
    state = state_type(states.StateContext(connection=connection))
    event = event_type(events.EventContext(connection=connection))

    next_state = state.on_event(event)

    assert isinstance(next_state, expected_next_state_type)

    if next_state is not state:
        # The new state should keep the event that led to it in its context.
        assert next_state.context.event is event


@pytest.mark.parametrize("state_type, event_type, expected_next_state_type", [
    (states.Disconnected, events.Up, states.Connecting),
    (states.Connecting, events.Connected, states.Connected),
    (states.Connected, events.Down, states.Disconnecting),
    (states.Disconnecting, events.Disconnected, states.Disconnected)
])
def test_happy_flow_state_transitions(state_type, event_type, expected_next_state_type):
    """
    {DISCONNECTED} --Up--> {CONNECTING} --Connected--> {CONNECTED}
    --Down--> {DISCONNECTING} --Disconnected--> {DISCONNECTED}
    """
    assert_state_transition(state_type, event_type, expected_next_state_type)


@pytest.mark.parametrize("event_type, expected_next_state_type", [
    (events.Up, states.Connecting),
    (events.Down, states.Disconnected),
    (events.Disconnected, states.Disconnected),
    (events.Connected, states. Disconnected),  # Invalid event.
    (events.UnexpectedError, states.Disconnected)  # Invalid event.
])
def test_disconnected_on_event_transitions(event_type, expected_next_state_type):
    assert_state_transition(states.Disconnected, event_type, expected_next_state_type)


@pytest.mark.parametrize("event_type, expected_next_state_type", [
    (events.Connected, states.Connected),
    (events.Down, states.Disconnecting),
    (events.UnexpectedError, states.Error),
    (events.Up, states.Disconnecting),  # Reconnection.
    (events.Disconnected, states.Disconnected)
])
def test_connecting_on_event_transitions(event_type, expected_next_state_type):
    assert_state_transition(states.Connecting, event_type, expected_next_state_type)


@pytest.mark.parametrize("event_type, expected_next_state_type", [
    (events.Down, states.Disconnecting),
    (events.Up, states.Disconnecting),  # Reconnection.
    (events.UnexpectedError, states.Error),
    (events.Disconnected, states.Disconnected),
    (events.Connected, states.Connected)
])
def test_connected_on_event_transitions(event_type, expected_next_state_type):
    assert_state_transition(states.Connected, event_type, expected_next_state_type)


@pytest.mark.parametrize("event_type, expected_next_state_type", [
    (events.Disconnected, states.Disconnected),
    (events.Up, states.Disconnecting),  # Reconnection.
    (events.Down, states.Disconnecting),
    (events.UnexpectedError, states.Disconnected),  # Errors events also signal VPN disconnection
    (events.Connected, states.Disconnecting)  # Invalid event.
])
def test_disconnecting_on_event_transitions(event_type, expected_next_state_type):
    assert_state_transition(states.Disconnecting, event_type, expected_next_state_type)


@pytest.mark.parametrize("event_type, expected_next_state_type", [
    (events.Down, states.Disconnected),
    (events.Up, states.Disconnecting),
    (events.UnexpectedError, states.Error),
    (events.Connected, states.Connected),
    (events.Disconnected, states.Error)  # Invalid event.
])
def test_error_on_event_transitions(event_type, expected_next_state_type):
    assert_state_transition(states.Error, event_type, expected_next_state_type)


@pytest.mark.parametrize("active_state_type", [
    states.Connecting, states.Connected, states.Disconnecting, states.Error
])
def test_reconnection_is_triggered_when_up_event_is_received_while_a_connection_is_active(
        active_state_type
):
    """
    A connection is active while in Connecting, Connected and Disconnecting
    states. When one of these states receives an Up event then a reconnection
    will be triggered. That means that, the current state will transition to
    Disconnecting state (to start disconnection) while keeping the new connection
    to be started (carried by the Up event) once the Disconnected state is reached.
    """
    active_state = active_state_type(states.StateContext(connection=Mock()))
    up = events.Up(events.EventContext(connection=Mock()))

    disconnecting = active_state.on_event(up)

    assert isinstance(disconnecting, states.Disconnecting)
    # The connection to disconnect from is the same we were connecting to.
    assert disconnecting.context.connection is active_state.context.connection
    # The connection that we want to reconnect to is the one carried by the up event.
    assert disconnecting.context.reconnection is up.context.connection


@pytest.mark.asyncio
async def test_disconnected_run_tasks_when_reconnection_is_not_requested_and_kill_switch_is_not_permanent():
    """
    When reconnection is not requested and the kill switch is not set to permanent,
    the disconnected state should run the following tasks:
     - Remove persisted connection parameters.
     - Disable kill switch.
     - Disable IPv6 leak protection.
    """
    context = Mock(spec=states.StateContext)
    context.reconnection = None  # Reconnection not requested
    context.kill_switch_setting = KillSwitchSetting.ON
    context.kill_switch.disable_ipv6_leak_protection = AsyncMock(return_value=None)
    context.kill_switch.disable = AsyncMock(return_value=None)
    context.connection.remove_persistence = AsyncMock(return_value=None)
    context.split_tunneling.clear_config = AsyncMock(return_value=None)

    disconnected = states.Disconnected(context=context)
    generated_event = await disconnected.run_tasks()

    assert context.method_calls == [
        call.connection.remove_persistence(),
        call.kill_switch.disable(),
        call.kill_switch.disable_ipv6_leak_protection(),
        call.split_tunneling.clear_config()
    ]

    assert generated_event is None


@pytest.mark.asyncio
async def test_disconnected_run_tasks_does_not_disable_the_kill_switch_when_set_to_permanent():
    """
    When the kill switch is not set to permanent, the disconnected state should
    **not** disable the kill switch.
    """
    context = AsyncMock()
    context.reconnection = None  # Reconnection not requested
    context.kill_switch_setting = KillSwitchSetting.PERMANENT
    context.connection.remove_persistence = AsyncMock(return_value=None)
    context.split_tunneling.clear_config = AsyncMock(return_value=None)

    disconnected = states.Disconnected(context=context)
    generated_event = await disconnected.run_tasks()

    assert context.method_calls == [
        call.connection.remove_persistence(),
        call.kill_switch.enable(permanent=True),
        call.split_tunneling.clear_config()
    ]

    assert generated_event is None


@pytest.mark.asyncio
async def test_disconnected_run_tasks_when_reconnection_is_requested_and_should_return_up_event():
    """
    When reconnection **is** requested while on the disconnected state then:
     - No connection tasks should be performed. It's very important that
       IPv6 leak protection or the kill switch are **not** disabled.AsyncMock
     - An Up event should be returned with the new connection to be started.
    """
    context = AsyncMock()
    context.reconnection = Mock()
    disconnected = states.Disconnected(context=context)

    generated_event = await disconnected.run_tasks()

    assert context.method_calls == [
        call.connection.remove_persistence(),
        call.kill_switch.enable()  # Kill switch is enabled to avoid leaks when switching servers.
    ]
    assert isinstance(generated_event, events.Up)
    assert generated_event.context.connection is context.reconnection


@pytest.mark.asyncio
async def test_disconnected_run_tasks_when_there_is_no_connection():
    """
    When there is no current connection and reconnection was not requested,
    the disconnect state should run the following taks:
     - disable the kill switch
     - disable IPv6 leak protection.
     """
    context = AsyncMock()
    context.connection = None
    context.reconnection = None
    disconnected = states.Disconnected(context=context)

    generated_event = await disconnected.run_tasks()

    assert context.method_calls == [
        call.kill_switch.disable(),
        call.kill_switch.disable_ipv6_leak_protection(),
        call.split_tunneling.clear_config()
    ]
    assert generated_event is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "kill_switch_setting", [KillSwitchSetting.ON, KillSwitchSetting.PERMANENT, KillSwitchSetting.OFF]
)
async def test_connecting_run_tasks(kill_switch_setting):
    """
    The connecting state tasks are the following ones, in the specified order:

     1. Enable IPv6 leak protection.
     2. Enable kill switch if it's set to be enabled.
     3. Start the connection.

    It's very important that IPv6 leak protection (and kill switch) is enabled
    before starting the connection.
    """
    context = AsyncMock()
    context.kill_switch_setting = kill_switch_setting

    connecting = states.Connecting(context=context)

    await connecting.run_tasks()

    permanent_ks = kill_switch_setting == KillSwitchSetting.PERMANENT
    assert context.method_calls == [
        call.kill_switch.enable(context.connection.server, permanent=permanent_ks),
        call.connection.start()
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "kill_switch_setting", [KillSwitchSetting.ON, KillSwitchSetting.PERMANENT, KillSwitchSetting.OFF]
)
async def test_connected_run_tasks(kill_switch_setting):
    """The tasks to be run while on the connected state is to persist the connection parameters and
    enable kill switch if it's set to be enabled."""
    context = AsyncMock(name="context")
    context.kill_switch_setting = kill_switch_setting
    context.event.context.forwarded_port = None
    context.split_tunneling_setting = Mock(name="split_tunneling_setting")

    connected = states.Connected(context)

    await connected.run_tasks()

    if kill_switch_setting == KillSwitchSetting.ON:
        assert context.method_calls == [
            call.kill_switch.enable(permanent=False),
            call.connection.add_persistence()
        ]
    elif kill_switch_setting == KillSwitchSetting.PERMANENT:
        assert context.method_calls == [
            call.kill_switch.enable(permanent=True),
            call.connection.add_persistence()
        ]
    else:  # Kill switch OFF.
        assert context.method_calls == [
            call.kill_switch.enable_ipv6_leak_protection(),
            call.kill_switch.disable(),
            call.split_tunneling.set_config(context.split_tunneling_setting.get_config()),
            call.connection.add_persistence(),
        ]


@pytest.mark.asyncio
async def test_disconnecting_run_tasks_stops_connection():
    """The only task be run while on the disconnecting state is to stop the connection."""
    connection = Mock()
    connection.stop = AsyncMock(return_value=None)
    disconnecting = states.Disconnecting(states.StateContext(connection=connection))

    await disconnecting.run_tasks()

    connection_calls = connection.method_calls
    assert len(connection_calls) == 1
    connection.stop.assert_called_once()


@pytest.mark.asyncio
async def test_connected_run_tasks_swallows_split_tunneling_errors():
    connection = Mock()
    connection.add_persistence = AsyncMock()
    context = states.StateContext(connection=connection)
    context.kill_switch = AsyncMock()
    context.kill_switch_setting = KillSwitchSetting.OFF
    context.split_tunneling = AsyncMock()
    context.split_tunneling_setting = SplitTunnelingSetting(enabled=True)
    context.split_tunneling.set_config = AsyncMock(side_effect=SplitTunnelingError("forced error"))
    connected = states.Connected(context)

    await connected.run_tasks()

    # The split tunneling exception shouldn't have bubbled up when applying ST config
    context.split_tunneling.set_config.assert_called_once()


@pytest.mark.asyncio
async def test_disconnected_run_tasks_swallows_split_tunneling_errors():
    context = states.StateContext(connection=None, reconnection=None)
    context.kill_switch = AsyncMock()
    context.kill_switch_setting = KillSwitchSetting.OFF
    context.split_tunneling = AsyncMock()
    context.split_tunneling_setting = SplitTunnelingSetting(enabled=True)
    context.split_tunneling.clear_config = AsyncMock(side_effect=SplitTunnelingError("forced error"))
    disconnected = states.Disconnected(context)

    await disconnected.run_tasks()

    # The split tunneling exception shouldn't have bubbled up when clearing ST config
    context.split_tunneling.clear_config.assert_called_once()


@pytest.mark.asyncio
async def test_error_state_run_tasks_persists_connection():
    """The only task to be run while in the error state is to persist the connection."""
    connection = AsyncMock()
    error_state = states.Error(states.StateContext(connection=connection))

    await error_state.run_tasks()

    connection_calls = connection.method_calls
    assert len(connection_calls) == 1
    connection.add_persistence.assert_called_once()



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from proton.vpn.session.exceptions import ServerNotFoundError
from proton.vpn.session.servers.types import PhysicalServer, LogicalServer, ServerLoad
from proton.vpn.session.servers.country_codes import get_country_name_by_code

import pytest


ID = "PHYS_SERVER_1"
ENTRY_IP = "192.168.0.1"
EXIT_IP = "192.168.0.2"
DOMAIN = "test.mock-domain.net"
STATUS = 1
GENERATION = 0
LABEL = "TestLabel"
SERVICEDOWNREASON = None
X25519_PK = "UBA8UbeQMmwfFeBp2lwwqwa/aF606BQKjzKHmNoJ03E="

MOCK_PHYSICAL = {
    "ID": ID,
    "EntryIP": ENTRY_IP,
    "ExitIP": EXIT_IP,
    "Domain": DOMAIN,
    "Status": STATUS,
    "Generation": GENERATION,
    "Label": LABEL,
    "ServicesDownReason": SERVICEDOWNREASON,
    "X25519PublicKey": X25519_PK,
}


NAME = "MOCK-SERVER#1"
ENTRYCOUNTRY = "CA"
EXITCOUNTRY = "CA"
TIER = 0
FEATURES = 0
STATE = None
CITY = "Toronto"
SCORE = 2.4273928
HOSTCOUNTRY = None
L_ID = "BzHqSTaqcpjIY9SncE5s7FpjBrPjiGOucCyJmwA6x4nTNqlElfKvCQFr9xUa2KgQxAiHv4oQQmAkcA56s3ZiGQ=="
LAT = 32
LONG = 40
L_STATUS = 1
LOAD = 45


MOCK_LOGICAL = {
    "Name": NAME,
    "EntryCountry": ENTRYCOUNTRY,
    "ExitCountry": EXITCOUNTRY,
    "Domain": DOMAIN,
    "Tier": TIER,
    "Features": FEATURES,
    "State": STATE,
    "City": CITY,
    "Score": SCORE,
    "HostCountry": HOSTCOUNTRY,
    "ID": L_ID,
    "Location": {
        "Lat": LAT, "Long": LONG
    },
    "Status": L_STATUS,
    "Servers": [MOCK_PHYSICAL],
    "Load": LOAD
}


class TestPhysicalServer:
    def test_init_server(self):
        server = PhysicalServer(MOCK_PHYSICAL)
        assert server.id == ID
        assert server.entry_ip == ENTRY_IP
        assert server.exit_ip == EXIT_IP
        assert server.domain == DOMAIN
        assert server.enabled == STATUS
        assert server.generation == GENERATION
        assert server.label == LABEL
        assert server.services_down_reason == SERVICEDOWNREASON
        assert server.x25519_pk == X25519_PK


class TestLogicalServer:
    def test_init_server(self):
        server = LogicalServer(MOCK_LOGICAL)
        assert server.id == L_ID
        assert server.load == LOAD
        assert server.score == SCORE
        assert server.enabled == L_STATUS
        assert server.name == NAME
        assert server.entry_country == ENTRYCOUNTRY
        assert server.entry_country_name == get_country_name_by_code(server.entry_country)
        assert server.exit_country == EXITCOUNTRY
        assert server.exit_country_name == get_country_name_by_code(server.exit_country)
        assert server.host_country == HOSTCOUNTRY
        assert server.features == []
        assert server.location == (STATE or CITY)
        assert server.city == CITY
        assert server.state == STATE
        assert server.tier == TIER
        assert server.latitude == LAT
        assert server.longitude == LONG
        assert server.physical_servers[0].domain == PhysicalServer(MOCK_PHYSICAL).domain
        assert server.physical_servers[0].entry_ip == PhysicalServer(MOCK_PHYSICAL).entry_ip
        assert server.physical_servers[0].exit_ip == PhysicalServer(MOCK_PHYSICAL).exit_ip

    def test_update(self):
        server = LogicalServer(MOCK_LOGICAL)

        server_load = ServerLoad({
            "ID": L_ID,
            "Load": 55,
            "Score": 3.14159,
            "enabled": 0
        })
        server.update(server_load)

        assert server.load == 55
        assert server.score == 3.14159
        assert not server.enabled

    def test_get_data(self):
        server = LogicalServer(MOCK_LOGICAL)
        _data = server.data
        _data["Name"] = "test-name"
        assert server.name != _data["Name"]

    def test_get_random_server(self):
        server = LogicalServer(MOCK_LOGICAL)
        _s = server.get_random_physical_server()
        assert _s.x25519_pk == X25519_PK

    def test_get_random_server_raises_exception(self):
        logical_copy = MOCK_LOGICAL.copy()
        logical_copy["Servers"] = []
        server = LogicalServer(logical_copy)

        with pytest.raises(ServerNotFoundError):
            server.get_random_physical_server()



"""
Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock, patch
import pytest
import time


from proton.vpn.session.feature_flags_fetcher import FeatureFlagsFetcher, FeatureFlags

EXPIRATION_TIME = time.time()


@pytest.fixture
def api_data():
    TEST_DATA = {
        "Code": 1000,
        "toggles": [
            {
                "name": "EnabledInApiDisabledInDefault",
                "enabled": True,
                "impressionData": False,
                "variant": {
                    "name": "disabled",
                    "enabled": False
                }
            },
            {
                "name": "DisabledInApiEnabledInDefault",
                "enabled": False,
                "impressionData": False,
                "variant": {
                    "name": "disabled",
                    "enabled": False
                }
            },
        ]
        }
    return TEST_DATA

@pytest.fixture
def default_data():
    TEST_DATA = {
        "toggles": [
            {
                "name": "EnabledInApiDisabledInDefault",
                "enabled": False,
                "impressionData": False,
                "variant": {
                    "name": "disabled",
                    "enabled": False
                }
            },
            {
                "name": "DisabledInApiEnabledInDefault",
                "enabled": True,
                "impressionData": False,
                "variant": {
                    "name": "disabled",
                    "enabled": False
                }
            }
        ]
        }
    return TEST_DATA


@patch("proton.vpn.session.feature_flags_fetcher.rest_api_request")
@pytest.mark.asyncio
async def test_fetch_returns_feature_flags_from_proton_rest_api(mock_rest_api_request, api_data):
    mock_cache_handler = Mock()
    mock_refresh_calculator = Mock()
    expiration_time_in_seconds = 10

    mock_refresh_calculator.get_expiration_time.return_value = expiration_time_in_seconds
    mock_rest_api_request.return_value = api_data

    ff = FeatureFlagsFetcher(Mock(), mock_refresh_calculator, mock_cache_handler)

    features = await ff.fetch()

    assert features.get("EnabledInApiDisabledInDefault") == True
    assert features.get("DisabledInApiEnabledInDefault") == False


def test_load_from_cache_returns_feature_flags_from_cache(api_data):
    mock_cache_handler = Mock()
    expiration_time_in_seconds = time.time()
    api_data["ExpirationTime"] = expiration_time_in_seconds

    mock_cache_handler.load.return_value = api_data

    ff = FeatureFlagsFetcher(Mock(), Mock(), mock_cache_handler)

    features = ff.load_from_cache()

    assert features.get("EnabledInApiDisabledInDefault") == True
    assert features.get("DisabledInApiEnabledInDefault") == False

@patch('proton.vpn.session.feature_flags_fetcher.FeatureFlags.default')
def test_load_from_cache_returns_default_feature_flags_when_no_cache_is_found(feature_flags_mock,default_data):
    feature_flags_mock.return_value = FeatureFlags(default_data)
    mock_cache_handler = Mock()
    mock_cache_handler.load.return_value = None
    ff = FeatureFlagsFetcher(Mock(), Mock(), mock_cache_handler)

    features = ff.load_from_cache()
    
    assert features.get("EnabledInApiDisabledInDefault") == False
    assert features.get("DisabledInApiEnabledInDefault") == True

def test_get_feature_flag_returns_false_when_feature_flag_does_not_exist(api_data):
    mock_cache_handler = Mock()
    mock_cache_handler.load.return_value = api_data
    ff = FeatureFlagsFetcher(Mock(), Mock(), mock_cache_handler)

    features = ff.load_from_cache()

    assert features.get("dummy-feature") is False
